import abc
import asyncio
import enum
import logging
from dataclasses import dataclass
from typing import cast, Tuple, Union, List, TYPE_CHECKING
from typing_extensions import Literal
from opentrons import types as top_types

if TYPE_CHECKING:
    from .dev_types import DoorStateNotificationType

MODULE_LOG = logging.getLogger(__name__)


class OutOfBoundsMove(RuntimeError):
    def __init__(self, message: str):
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        return f"OutOfBoundsMove: {self.message}"

    def __repr__(self) -> str:
        return f"<{str(self.__class__)}: {self.message}>"


class MotionChecks(enum.Enum):
    NONE = 0
    LOW = 1
    HIGH = 2
    BOTH = 3


class Axis(enum.Enum):
    X = 0
    Y = 1
    Z = 2
    A = 3
    B = 4
    C = 5

    @classmethod
    def by_mount(cls, mount: top_types.Mount):
        bm = {top_types.Mount.LEFT: cls.Z, top_types.Mount.RIGHT: cls.A}
        return bm[mount]

    @classmethod
    def gantry_axes(cls) -> Tuple["Axis", "Axis", "Axis", "Axis"]:
        """The axes which are tied to the gantry and require the deck
        calibration transform
        """
        return (cls.X, cls.Y, cls.Z, cls.A)

    @classmethod
    def of_plunger(cls, mount: top_types.Mount):
        pm = {top_types.Mount.LEFT: cls.B, top_types.Mount.RIGHT: cls.C}
        return pm[mount]

    @classmethod
    def to_mount(cls, inst: "Axis"):
        return {
            cls.Z: top_types.Mount.LEFT,
            cls.A: top_types.Mount.RIGHT,
            cls.B: top_types.Mount.LEFT,
            cls.C: top_types.Mount.RIGHT,
        }[inst]

    def __str__(self):
        return self.name


class DoorState(enum.Enum):
    OPEN = False
    CLOSED = True

    def __str__(self):
        return self.name.lower()


class HardwareEventType(enum.Enum):
    DOOR_SWITCH_CHANGE = enum.auto()


@dataclass
class DoorStateNotification:
    event: "DoorStateNotificationType" = HardwareEventType.DOOR_SWITCH_CHANGE
    new_state: DoorState = DoorState.CLOSED
    blocking: bool = False


# new event types get new dataclasses
# when we add more event types we add them here
HardwareEvent = Union[DoorStateNotification]


class HardwareAPILike(abc.ABC):
    """A dummy class useful in isinstance checks to accept an API or adapter"""

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        ...

    @property
    def board_revision(self) -> str:
        ...

    @property
    def door_state(self) -> DoorState:
        ...

    @door_state.setter
    def door_state(self, door_state: DoorState) -> DoorState:
        ...

    def validate_calibration(self):
        ...


RevisionLiteral = Literal["2.1", "A", "B", "C", "UNKNOWN"]


class BoardRevision(enum.Enum):
    UNKNOWN = enum.auto()
    OG = enum.auto()
    A = enum.auto()
    B = enum.auto()
    C = enum.auto()

    @classmethod
    def by_bits(cls, rev_bits: Tuple[bool, bool]):
        br = {
            (True, True): cls.OG,
            (False, True): cls.A,
            (True, False): cls.B,
            (False, False): cls.C,
        }
        return br[rev_bits]

    def real_name(self) -> Union[RevisionLiteral, Literal["UNKNOWN"]]:
        rn = "2.1" if self.name == "OG" else self.name
        return cast(Union[RevisionLiteral, Literal["UNKNOWN"]], rn)

    def __str__(self) -> str:
        return self.real_name()


class CriticalPoint(enum.Enum):
    """Possibilities for the point to move in a move call.

    The active critical point determines the offsets that are added to the
    gantry position when moving a pipette around.
    """

    MOUNT = enum.auto()
    """
    For legacy reasons, the position of the end of a P300 single. The default
    when no pipette is attached, and used for consistent behavior in certain
    contexts (like change pipette) when a variety of different pipettes might
    be attached.
    """

    NOZZLE = enum.auto()
    """
    The end of the nozzle of a single pipette or the end of the back-most
    nozzle of a multipipette. Only relevant when a pipette is present.
    """

    TIP = enum.auto()
    """
    The end of the tip of a single pipette or the end of the back-most
    tip of a multipipette. Only relevant when a pipette is present and
    a tip with known tip length is attached.
    """

    XY_CENTER = enum.auto()
    """
    Separately from the z component of the critical point, XY_CENTER means
    the critical point under consideration is the XY center of the pipette.
    This changes nothing for single pipettes, but makes multipipettes
    move their centers - so between channels 4 and 5 - to the specified
    point.
    """

    FRONT_NOZZLE = enum.auto()
    """
    The end of the front-most nozzle of a multipipette with a tip attached.
    Only relevant when a multichannel pipette is present.
    """


class ExecutionState(enum.Enum):
    RUNNING = enum.auto()
    PAUSED = enum.auto()
    CANCELLED = enum.auto()

    def __str__(self):
        return self.name


class PipettePair(enum.Enum):
    PRIMARY_LEFT = enum.auto()
    PRIMARY_RIGHT = enum.auto()

    @property
    def primary(self) -> top_types.Mount:
        if self.name == "PRIMARY_RIGHT":
            return top_types.Mount.RIGHT
        else:
            return top_types.Mount.LEFT

    @property
    def secondary(self) -> top_types.Mount:
        if self.name == "PRIMARY_RIGHT":
            return top_types.Mount.LEFT
        else:
            return top_types.Mount.RIGHT

    @classmethod
    def of_mount(cls, mount: top_types.Mount) -> "PipettePair":
        pair = {
            top_types.Mount.LEFT: cls.PRIMARY_LEFT,
            top_types.Mount.RIGHT: cls.PRIMARY_RIGHT,
        }
        return pair[mount]


class HardwareAction(enum.Enum):
    DROPTIP = enum.auto()
    ASPIRATE = enum.auto()
    DISPENSE = enum.auto()
    BLOWOUT = enum.auto()
    PREPARE_ASPIRATE = enum.auto()

    def __str__(self):
        return self.name


class PauseType(enum.Enum):
    PAUSE = 0
    DELAY = 1


@dataclass
class AionotifyEvent:
    flags: enum.EnumMeta
    name: str

    @classmethod
    def build(cls, name: str, flags: List[enum.Enum]) -> "AionotifyEvent":
        # See https://github.com/python/mypy/issues/5317
        # as to why mypy cannot detect that list
        # comprehension or variables cannot be dynamically
        # determined to meet the argument criteria for
        # enums. Hence, the type ignore below.
        flag_list = [f.name for f in flags]
        Flag = enum.Enum("Flag", flag_list)  # type: ignore
        return cls(flags=Flag, name=name)


class PauseResumeError(RuntimeError):
    pass


class ExecutionCancelledError(RuntimeError):
    pass


class MustHomeError(RuntimeError):
    pass


class NoTipAttachedError(RuntimeError):
    pass


class TipAttachedError(RuntimeError):
    pass


class PairedPipetteConfigValueError(RuntimeError):
    pass



"""JSON file reading."""

from opentrons.protocols.models import JsonProtocol
from .protocol_file import ProtocolFile


class JsonFileReader:
    """Reads and parses JSON protocol files."""

    @staticmethod
    def read(protocol_file: ProtocolFile) -> JsonProtocol:
        """Read and parse file into a JsonProtocol model."""
        # TODO(mc, 2021-08-25): validate files list length before access
        contents = protocol_file.files[0].read_text(encoding="utf-8")
        return JsonProtocol.parse_raw(contents)



"""Protocol run control and management."""
from dataclasses import dataclass
from typing import List, Optional

from opentrons.protocol_engine import (
    ProtocolEngine,
    Command,
    LoadedLabware,
    LoadedPipette,
)

from .protocol_file import ProtocolFile, ProtocolFileType
from .task_queue import TaskQueue, TaskQueuePhase
from .json_file_reader import JsonFileReader
from .json_command_translator import JsonCommandTranslator
from .python_file_reader import PythonFileReader
from .python_context_creator import PythonContextCreator
from .python_executor import PythonExecutor


@dataclass(frozen=True)
class ProtocolRunData:
    """Data from a protocol run."""

    commands: List[Command]
    labware: List[LoadedLabware]
    pipettes: List[LoadedPipette]


class ProtocolRunner:
    """An interface to manage and control a protocol run.

    The ProtocolRunner is primarily responsible for feeding a ProtocolEngine
    with commands and control signals. These commands and signals are
    generated by protocol files, hardware signals, or externally via
    the HTTP robot-server.

    A ProtocolRunner controls a single run. Once the run is finished,
    you will need a new ProtocolRunner to do another run.
    """

    def __init__(
        self,
        protocol_engine: ProtocolEngine,
        task_queue: Optional[TaskQueue] = None,
        json_file_reader: Optional[JsonFileReader] = None,
        json_command_translator: Optional[JsonCommandTranslator] = None,
        python_file_reader: Optional[PythonFileReader] = None,
        python_context_creator: Optional[PythonContextCreator] = None,
        python_executor: Optional[PythonExecutor] = None,
    ) -> None:
        """Initialize the ProtocolRunner with its dependencies."""
        self._protocol_engine = protocol_engine
        self._task_queue = task_queue or TaskQueue()
        self._json_file_reader = json_file_reader or JsonFileReader()
        self._json_command_translator = (
            json_command_translator or JsonCommandTranslator()
        )
        self._python_file_reader = python_file_reader or PythonFileReader()
        self._python_context_creator = python_context_creator or PythonContextCreator()
        self._python_executor = python_executor or PythonExecutor()

    def load(self, protocol_file: ProtocolFile) -> None:
        """Load a ProtocolFile into managed ProtocolEngine.

        Calling this method is only necessary if the runner will be used
        to control the run of a file-based protocol.
        """
        protocol_type = protocol_file.protocol_type

        if protocol_type == ProtocolFileType.JSON:
            self._load_json(protocol_file)

        elif protocol_type == ProtocolFileType.PYTHON:
            self._load_python(protocol_file)

    def play(self) -> None:
        """Start or resume the run."""
        self._protocol_engine.play()

        if not self._task_queue.is_started():
            self._task_queue.add(
                phase=TaskQueuePhase.CLEANUP,
                func=self._protocol_engine.stop,
                wait_until_complete=True,
            )
            self._task_queue.start()

    def pause(self) -> None:
        """Pause the run."""
        self._protocol_engine.pause()

    async def stop(self) -> None:
        """Stop (cancel) the run."""
        await self._protocol_engine.halt()

    async def join(self) -> None:
        """Wait for the run to complete, propagating any errors.

        This method may be called before the run starts, in which case,
        it will wait for the run to start before waiting for completion.
        """
        return await self._task_queue.join()

    async def run(self, protocol_file: ProtocolFile) -> ProtocolRunData:
        """Run a given protocol to completion."""
        self.load(protocol_file)
        self.play()
        await self.join()

        commands = self._protocol_engine.state_view.commands.get_all()
        labware = self._protocol_engine.state_view.labware.get_all()
        pipettes = self._protocol_engine.state_view.pipettes.get_all()

        return ProtocolRunData(commands=commands, labware=labware, pipettes=pipettes)

    def _load_json(self, protocol_file: ProtocolFile) -> None:
        protocol = self._json_file_reader.read(protocol_file)
        commands = self._json_command_translator.translate(protocol)
        for request in commands:
            self._protocol_engine.add_command(request=request)

    def _load_python(self, protocol_file: ProtocolFile) -> None:
        protocol = self._python_file_reader.read(protocol_file)
        context = self._python_context_creator.create(self._protocol_engine)
        self._task_queue.add(
            phase=TaskQueuePhase.RUN,
            func=self._python_executor.execute,
            protocol=protocol,
            context=context,
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, cast

from opentrons import types
from opentrons.hardware_control.modules import Thermocycler
from opentrons.hardware_control.types import CriticalPoint
from opentrons.protocol_api.labware import Labware, Well
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.context.paired_instrument import AbstractPairedInstrument
from opentrons.protocols.context.protocol import AbstractProtocol
from opentrons.protocols.geometry import planning
from opentrons.protocols.api_support.util import build_edges
from opentrons.protocols.geometry.module_geometry import ThermocyclerGeometry

if TYPE_CHECKING:
    from opentrons.hardware_control import types as hc_types
    from opentrons.protocols.api_support.util import HardwareManager

logger = logging.getLogger(__name__)


class PairedInstrument(AbstractPairedInstrument):
    def __init__(
        self,
        primary_instrument: AbstractInstrument,
        secondary_instrument: AbstractInstrument,
        pair_policy: hc_types.PipettePair,
        ctx: AbstractProtocol,
        hardware_manager: HardwareManager,
    ):
        self.p_instrument = primary_instrument
        self.s_instrument = secondary_instrument
        self._pair_policy = pair_policy
        self._ctx = ctx
        self._hw_manager = hardware_manager

    def pick_up_tip(
        self,
        target: Well,
        secondary_target: Well,
        tiprack: Labware,
        presses: Optional[int],
        increment: Optional[float],
        tip_length: float,
    ):
        self.move_to(target.top())

        self._hw_manager.hardware.set_current_tiprack_diameter(
            self._pair_policy, target.diameter
        )
        self._hw_manager.hardware.pick_up_tip(
            self._pair_policy, tip_length, presses, increment
        )

        self._hw_manager.hardware.set_working_volume(
            self._pair_policy, target.max_volume
        )

        tiprack.use_tips(target, self.p_instrument.get_channels())
        tiprack.use_tips(secondary_target, self.s_instrument.get_channels())

    def drop_tip(self, target: types.Location, home_after: bool):
        self.move_to(target)
        self._hw_manager.hardware.drop_tip(self._pair_policy, home_after=home_after)
        return self

    def move_to(
        self,
        location: types.Location,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ):
        if not speed:
            speed = self.p_instrument.get_default_speed()

        last_location = self._ctx.get_last_location()
        if last_location:
            from_lw = last_location.labware
        else:
            from_lw = LabwareLike(None)

        from_center = "centerMultichannelOnWells" in from_lw.quirks_from_any_parent()
        cp_override = CriticalPoint.XY_CENTER if from_center else None
        from_loc = types.Location(
            self._hw_manager.hardware.gantry_position(
                self._pair_policy.primary, critical_point=cp_override
            ),
            from_lw,
        )

        for mod in self._ctx.get_loaded_modules().values():
            if isinstance(mod.module, Thermocycler):
                cast(ThermocyclerGeometry, mod.geometry).flag_unsafe_move(
                    to_loc=location,
                    from_loc=from_loc,
                    lid_position=mod.module.lid_status,
                )

        primary_height = self._hw_manager.hardware.get_instrument_max_height(
            self._pair_policy.primary
        )
        secondary_height = self._hw_manager.hardware.get_instrument_max_height(
            self._pair_policy.secondary
        )
        moves = planning.plan_moves(
            from_loc,
            location,
            self._ctx.get_deck(),
            min(primary_height, secondary_height),
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
        )
        logger.debug("move_to: {}->{} via:\n\t{}".format(from_loc, location, moves))
        try:
            for move in moves:
                self._hw_manager.hardware.move_to(
                    self._pair_policy,
                    move[0],
                    critical_point=move[1],
                    speed=speed,
                    max_speeds=self._ctx.get_max_speeds().data,
                )
        except Exception:
            self._ctx.set_last_location(None)
            raise
        else:
            self._ctx.set_last_location(location)
        return self

    def aspirate(
        self,
        volume: float,
        location: types.Location,
        rate: float,
    ) -> None:

        if self.p_instrument.get_current_volume() == 0:
            # Make sure we're at the top of the labware and clear of any
            # liquid to prepare the pipette for aspiration
            primary_ready = self.p_instrument.get_pipette()["ready_to_aspirate"]
            secondary_ready = self.s_instrument.get_pipette()["ready_to_aspirate"]
            if not primary_ready or not secondary_ready:
                if location.labware.is_well:
                    self.move_to(location.labware.as_well().top())
                else:
                    # TODO(seth,2019/7/29): This should be a warning exposed
                    #  via rpc to the runapp
                    logger.warning(
                        "When aspirate is called on something other than a "
                        "well relative position, we can't move to the top of"
                        " the well to prepare for aspiration. This might "
                        "cause over aspiration if the previous command is a "
                        "blow_out."
                    )
                self._hw_manager.hardware.prepare_for_aspirate(self._pair_policy)
            self.move_to(location)
        elif location != self._ctx.get_last_location():
            self.move_to(location)

        self._hw_manager.hardware.aspirate(self._pair_policy, volume, rate)

    def dispense(self, volume: float, location: types.Location, rate: float) -> None:
        if location != self._ctx.get_last_location():
            self.move_to(location)
        self._hw_manager.hardware.dispense(self._pair_policy, volume, rate)

    def blow_out(self, location: types.Location):
        if location != self._ctx.get_last_location():
            self.move_to(location)
        self._hw_manager.hardware.blow_out(self._pair_policy)

    def touch_tip(self, well: Well, radius: float, v_offset: float, speed: float):
        edges = build_edges(
            well,
            v_offset,
            self._pair_policy.primary,
            self._ctx.get_deck(),
            radius,
        )
        for edge in edges:
            self._hw_manager.hardware.move_to(self._pair_policy, edge, speed)



import functools
import inspect
from typing import Any, Callable, Dict, Generic, Mapping, Optional, Tuple, TypeVar, cast
from opentrons.broker import Broker
from . import types as command_types


CacheT = TypeVar("CacheT")


class InspectMemoizer(Generic[CacheT]):
    def __init__(self, method: Callable[[Any], CacheT]) -> None:
        self._method = method
        self._cache: Dict[Callable, CacheT] = {}

    def get(self, f: Callable) -> CacheT:
        v = self._cache.get(f)
        if not v:
            v = self._method(f)
            self._cache[f] = v
        return v


signature_cache = InspectMemoizer(inspect.signature)
"""Cache inspect.signature method calls"""
getfullargspec_cache = InspectMemoizer(inspect.getfullargspec)
"""Cache inspect.getfullargspec method calls"""


class CommandPublisher:
    def __init__(self, broker: Optional[Broker]) -> None:
        self._broker = broker or Broker()

    @property
    def broker(self) -> Broker:
        return self._broker

    @broker.setter
    def broker(self, broker: Broker) -> None:
        self._broker = broker


CmdFunction = Callable[..., command_types.Command]


def do_publish(
    broker: Broker,
    cmd: CmdFunction,
    f: Callable,
    when: command_types.MessageSequenceId,
    res: Any,
    meta: Any,
    *args: Any,
    **kwargs: Any
) -> None:
    """Implement the publish so it can be called outside the decorator"""
    publish_command = functools.partial(broker.publish, topic=command_types.COMMAND)
    call_args = _get_args(f, args, kwargs)
    if when == "before":
        broker.logger.info(
            "{}: {}".format(
                f.__qualname__, {k: v for k, v in call_args.items() if str(k) != "self"}
            )
        )
    getfullargspec = getfullargspec_cache.get(cmd)
    command_args = dict(
        zip(reversed(getfullargspec.args), reversed(getfullargspec.defaults or []))
    )

    # TODO (artyom, 20170927): we are doing this to be able to use
    # the decorator in Instrument class methods, in which case
    # self is effectively an instrument.
    # To narrow the scope of this hack, we are checking if the
    # command is expecting instrument first.
    if "instrument" in getfullargspec.args:
        # We are also checking if call arguments have 'self' and
        # don't have instruments specified, in which case
        # instruments should take precedence.
        if "instrument" not in call_args and "self" in call_args:
            call_args["instrument"] = call_args["self"]

    command_args.update(
        {key: call_args[key] for key in (set(getfullargspec.args) & call_args.keys())}
    )

    if meta:
        command_args["meta"] = meta

    payload = cmd(**command_args)

    publish_command(message={**payload, "$": when})


def publish_paired(
    broker: Broker,
    cmd: CmdFunction,
    when: command_types.MessageSequenceId,
    res: Any,
    *args: Any,
    pub_type: str = "Paired Pipettes"
) -> None:
    """Implement a second publisher outside of the decorator that
    relies on the method providing all of the arguments required
    rather than binding defaults to the signature"""
    publish_command = functools.partial(broker.publish, topic=command_types.COMMAND)

    payload = cmd(*args, pub_type)

    publish_command(message={**payload, "$": when})


FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def publish(command: CmdFunction, meta: Any = None) -> Callable[[FuncT], FuncT]:
    """Publish events both before and after the decorated function has run."""

    def _decorator(f: FuncT) -> FuncT:
        @functools.wraps(
            f, updated=functools.WRAPPER_UPDATES + ("__globals__",)  # type: ignore[operator]  # noqa: E501
        )
        def _decorated(*args: Any, **kwargs: Any) -> Any:
            try:
                broker = cast(Broker, args[0].broker)
            except AttributeError:
                raise RuntimeError(
                    "Only methods of CommandPublisher \
                    classes should be decorated."
                )

            do_publish(broker, command, f, "before", None, meta, *args, **kwargs)
            res = f(*args, **kwargs)
            do_publish(broker, command, f, "after", res, meta, *args, **kwargs)
            return res

        return cast(FuncT, _decorated)

    return _decorator


def _get_args(f: Callable, args: Tuple, kwargs: Mapping[str, Any]) -> Dict[str, Any]:
    # Create the initial dictionary with args that have defaults
    res = {}
    sig = signature_cache.get(f)
    ismethod = inspect.ismethod(f)
    if ismethod and args[0] is f.__self__:  # type: ignore
        args = args[1:]
    if ismethod:
        res["self"] = f.__self__  # type: ignore

    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    res.update(bound.arguments)
    return res



from typing import Optional, cast

from opentrons import types
from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_api.labware import Well, Labware
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.context.paired_instrument import AbstractPairedInstrument
from opentrons.protocols.context.protocol import AbstractProtocol
from opentrons.protocols.geometry.module_geometry import ThermocyclerGeometry


class PairedInstrumentSimulation(AbstractPairedInstrument):
    def __init__(
        self,
        primary_instrument: AbstractInstrument,
        secondary_instrument: AbstractInstrument,
        ctx: AbstractProtocol,
    ):
        self._primary = primary_instrument
        self._secondary = secondary_instrument
        self._protocol_interface = ctx

    def pick_up_tip(
        self,
        target: Well,
        secondary_target: Well,
        tiprack: Labware,
        presses: Optional[int],
        increment: Optional[float],
        tip_length: float,
    ) -> None:
        self.move_to(target.top())

        self._primary.pick_up_tip(
            well=target._impl,
            tip_length=tip_length,
            presses=presses,
            increment=increment,
        )
        self._secondary.pick_up_tip(
            well=secondary_target._impl,
            tip_length=tip_length,
            presses=presses,
            increment=increment,
        )

        tiprack.use_tips(target, self._primary.get_channels())
        tiprack.use_tips(secondary_target, self._secondary.get_channels())

    def drop_tip(self, target: types.Location, home_after: bool) -> None:
        self._primary.drop_tip(home_after=home_after)
        self._secondary.drop_tip(home_after=home_after)

    def move_to(
        self,
        location: types.Location,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> None:
        # Check for unsafe moves while thermocycler is open.
        from_loc = self._protocol_interface.get_last_location()
        if not from_loc:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        for mod in self._protocol_interface.get_loaded_modules().values():
            if isinstance(mod.module, Thermocycler):
                cast(ThermocyclerGeometry, mod.geometry).flag_unsafe_move(
                    to_loc=location,
                    from_loc=from_loc,
                    lid_position=mod.module.lid_status,
                )

        self._primary.move_to(
            location=location,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
            speed=speed,
        )
        self._secondary.move_to(
            location=location,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
            speed=speed,
        )

    def aspirate(
        self,
        volume: float,
        location: types.Location,
        rate: float,
    ) -> None:
        self._primary.aspirate(volume, rate)
        self._secondary.aspirate(volume, rate)

    def dispense(self, volume: float, location: types.Location, rate: float) -> None:
        self._primary.dispense(volume, rate)
        self._secondary.dispense(volume, rate)

    def blow_out(self, location: types.Location) -> None:
        self._primary.blow_out()
        self._secondary.blow_out()

    def touch_tip(
        self, well: Well, radius: float, v_offset: float, speed: float
    ) -> None:
        pass



"""Protocol engine command models.

The `protocol_engine.commands` module contains the data models of all
command requests and results the engine is able to handle. It also contains
models and type definitions for command resources, where a command:

- May be pending, running, completed, or failed
- Contains request and, if applicable, result data
- Has timestamps for when it was created, started, completed, and/or failed

Request and result models in the module are defined using Pydantic, because
they are part of the public input / output of the engine, and need validation
and/or schema generation.
"""

from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest, CommandStatus
from .command_mapper import CommandMapper
from .command_unions import Command, CommandRequest, CommandResult, CommandType

from .add_labware_definition import (
    AddLabwareDefinition,
    AddLabwareDefinitionData,
    AddLabwareDefinitionRequest,
    AddLabwareDefinitionResult,
    AddLabwareDefinitionCommandType,
)

from .aspirate import (
    Aspirate,
    AspirateData,
    AspirateRequest,
    AspirateResult,
    AspirateCommandType,
)

from .dispense import (
    Dispense,
    DispenseData,
    DispenseRequest,
    DispenseResult,
    DispenseCommandType,
)

from .drop_tip import (
    DropTip,
    DropTipData,
    DropTipRequest,
    DropTipResult,
    DropTipCommandType,
)

from .load_labware import (
    LoadLabware,
    LoadLabwareData,
    LoadLabwareRequest,
    LoadLabwareResult,
    LoadLabwareCommandType,
)

from .load_pipette import (
    LoadPipette,
    LoadPipetteData,
    LoadPipetteRequest,
    LoadPipetteResult,
    LoadPipetteCommandType,
)

from .move_to_well import (
    MoveToWell,
    MoveToWellData,
    MoveToWellRequest,
    MoveToWellResult,
    MoveToWellCommandType,
)

from .pick_up_tip import (
    PickUpTip,
    PickUpTipData,
    PickUpTipRequest,
    PickUpTipResult,
    PickUpTipCommandType,
)

from .pause import (
    Pause,
    PauseData,
    PauseRequest,
    PauseResult,
    PauseCommandType,
)


__all__ = [
    # command model factory
    "CommandMapper",
    # command type unions
    "Command",
    "CommandRequest",
    "CommandResult",
    "CommandType",
    # base interfaces
    "AbstractCommandImpl",
    "BaseCommand",
    "BaseCommandRequest",
    "CommandStatus",
    # load labware command models
    "LoadLabware",
    "LoadLabwareRequest",
    "LoadLabwareData",
    "LoadLabwareResult",
    "LoadLabwareCommandType",
    # add labware definition command models
    "AddLabwareDefinition",
    "AddLabwareDefinitionRequest",
    "AddLabwareDefinitionData",
    "AddLabwareDefinitionResult",
    "AddLabwareDefinitionCommandType",
    # load pipette command models
    "LoadPipette",
    "LoadPipetteRequest",
    "LoadPipetteData",
    "LoadPipetteResult",
    "LoadPipetteCommandType",
    # move to well command models
    "MoveToWell",
    "MoveToWellRequest",
    "MoveToWellData",
    "MoveToWellResult",
    "MoveToWellCommandType",
    # pick up tip command models
    "PickUpTip",
    "PickUpTipRequest",
    "PickUpTipData",
    "PickUpTipResult",
    "PickUpTipCommandType",
    # drop tip command models
    "DropTip",
    "DropTipRequest",
    "DropTipData",
    "DropTipResult",
    "DropTipCommandType",
    # aspirate command models
    "Aspirate",
    "AspirateRequest",
    "AspirateData",
    "AspirateResult",
    "AspirateCommandType",
    # dispense command models
    "Dispense",
    "DispenseRequest",
    "DispenseData",
    "DispenseResult",
    "DispenseCommandType",
    # pause command models
    "Pause",
    "PauseData",
    "PauseRequest",
    "PauseResult",
    "PauseCommandType",
]



"""Errors raised by serial connection."""


class SerialException(Exception):
    """Base serial exception"""

    def __init__(self, port: str, description: str):
        super().__init__(f"{port}: {description}")
        self.port = port
        self.description = description


class NoResponse(SerialException):
    def __init__(self, port: str, command: str):
        super().__init__(port=port, description=f"No response to '{command}'")
        self.command = command


class FailedCommand(SerialException):
    def __init__(self, port: str, response: str):
        super().__init__(
            port=port, description=f"'Received error response '{response}'"
        )
        self.response = response


class AlarmResponse(FailedCommand):
    pass


class ErrorResponse(FailedCommand):
    pass



"""Add labware command."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type
from typing_extensions import Literal

from opentrons.protocols.models import LabwareDefinition
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest

AddLabwareDefinitionCommandType = Literal["addLabwareDefinition"]


class AddLabwareDefinitionData(BaseModel):
    """Data required to add a labware definition."""

    definition: LabwareDefinition = Field(..., description="The labware definition.")


class AddLabwareDefinitionResult(BaseModel):
    """Result of add labware request."""

    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The labware definition version.",
    )


class AddLabwareDefinitionImplementation(
    AbstractCommandImpl[AddLabwareDefinitionData, AddLabwareDefinitionResult]
):
    """Add labware command implementation."""

    async def execute(
        self, data: AddLabwareDefinitionData
    ) -> AddLabwareDefinitionResult:
        """Execute the add labware definition request."""
        return AddLabwareDefinitionResult(
            loadName=data.definition.parameters.loadName,
            namespace=data.definition.namespace,
            version=data.definition.version,
        )


class AddLabwareDefinition(
    BaseCommand[AddLabwareDefinitionData, AddLabwareDefinitionResult]
):
    """Add labware definition command resource."""

    commandType: AddLabwareDefinitionCommandType = "addLabwareDefinition"
    data: AddLabwareDefinitionData
    result: Optional[AddLabwareDefinitionResult]

    _ImplementationCls: Type[
        AddLabwareDefinitionImplementation
    ] = AddLabwareDefinitionImplementation


class AddLabwareDefinitionRequest(BaseCommandRequest[AddLabwareDefinitionData]):
    """Add labware definition command creation request."""

    commandType: AddLabwareDefinitionCommandType = "addLabwareDefinition"
    data: AddLabwareDefinitionData

    _CommandCls: Type[AddLabwareDefinition] = AddLabwareDefinition



""" opentrons.calibration_storage.file_operators: functions that
manipulate the file system.

These methods should only be imported inside the calibration_storage
module, except in the special case of v2 labware support in
the v1 API.
"""
import json
import datetime
import typing

from .types import StrPath
from .encoder_decoder import DateTimeEncoder, DateTimeDecoder


DecoderType = typing.Type[json.JSONDecoder]
EncoderType = typing.Type[json.JSONEncoder]


def read_cal_file(
    filepath: StrPath, decoder: DecoderType = DateTimeDecoder
) -> typing.Dict:
    """
    Function used to read data from a file

    :param filepath: path to look for data at
    :param decoder: if there is any specialized decoder needed.
    The default decoder is the date time decoder.
    :return: Data from the file
    """
    # TODO(6/16): We should use tagged unions for
    # both the calibration and tip length dicts to better
    # categorize the Typed Dicts used here.
    # This can be done when the labware endpoints
    # are refactored to grab tip length calibration
    # from the correct locations.
    with open(filepath, "r") as f:
        calibration_data = json.load(f, cls=decoder)
    if isinstance(calibration_data.values(), dict):
        for value in calibration_data.values():
            if value.get("lastModified"):
                assert isinstance(value["lastModified"], datetime.datetime), (
                    "invalid decoded value type for lastModified: got "
                    f"{type(value['lastModified']).__name__},"
                    "expected datetime"
                )
    return calibration_data


def save_to_file(
    filepath: StrPath, data: typing.Mapping, encoder: EncoderType = DateTimeEncoder
):
    """
    Function used to save data to a file

    :param filepath: path to save data at
    :param data: data to save
    :param encoder: if there is any specialized encoder needed.
    The default encoder is the date time encoder.
    """
    with open(filepath, "w") as f:
        json.dump(data, f, cls=encoder)



"""
USB Driver.

A class to convert info from the usb bus into a
more readable format.
"""

import subprocess
import re
import os
from typing import List, Set

from opentrons.algorithms.dfs import DFS
from opentrons.hardware_control.modules.types import ModuleAtPort
from opentrons.hardware_control.types import BoardRevision

from .interfaces import USBDriverInterface
from .types import USBPort


# Example usb path might look like:
# '/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3:1.0/tty/ttyACM1/dev'.
# There is only 1 bus that supports USB on the raspberry pi.
BUS_PATH = "/sys/bus/usb/devices/usb1/"
PORT_PATTERN = r"(/\d-\d(\.?\d)+)+:"
DEVICE_PATH = r"\d.\d/tty/tty(\w{4})/dev"
USB_PORT_INFO = re.compile(PORT_PATTERN + DEVICE_PATH)


class USBBus(USBDriverInterface):
    def __init__(self, board_revision: BoardRevision):
        self._board_revision = board_revision
        self._usb_dev: List[USBPort] = self.read_usb_bus()
        self._dfs: DFS = DFS(self._usb_dev)
        self._sorted = self._dfs.dfs()

    @staticmethod
    def read_bus() -> List[str]:
        """
        Read the USB Bus information.

        Use the sys bus path to find all of the USBs with
        active devices connected to them.
        """
        read = [""]
        try:
            read = (
                subprocess.check_output(["find", BUS_PATH, "-name", "dev"])
                .decode()
                .splitlines()
            )
        except Exception:
            pass
        return read

    @staticmethod
    def read_symlink(virtual_port: str) -> str:
        """ """
        symlink = ""
        try:
            symlink = os.readlink(virtual_port)
        except OSError:
            pass
        return symlink

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    @property
    def usb_dev(self) -> List[USBPort]:
        """
        USBBus property: usb_dev.

        :returns: The list of ports found from
        the usb bus.
        """
        return self._usb_dev

    @usb_dev.setter
    def usb_dev(self, ports: List[USBPort]) -> None:
        """
        USBBus setter: usb_dev.

        :param ports: The list of ports found from
        the usb bus.
        """
        self._usb_dev = ports

    @property
    def sorted_ports(self) -> Set:
        """
        USBBus property: sorted_ports.

        :returns: The set of sorted ports
        """
        return self._sorted

    @sorted_ports.setter
    def sorted_ports(self, sorted: Set) -> None:
        """
        USBBus setter: sorted_ports.

        :param sorted: The updated set of usb ports.
        """
        self._sorted = sorted

    def read_usb_bus(self) -> List[USBPort]:
        """
        Read usb bus

        Take the value returned from the USB bus and match
        the paths to the expected port paths for modules.
        :returns: A list of matching ports as dataclasses
        """
        active_ports = self.read_bus()
        port_matches = []
        for port in active_ports:
            match = USB_PORT_INFO.search(port)
            if match:
                port_matches.append(
                    USBPort.build(match.group(0).strip("/"), self.board_revision)
                )
        return port_matches

    def find_port(self, device_path: str) -> USBPort:
        """
        Find port.

        Take the value returned from the USB bus and match
        the paths to the expected port paths for modules.
        :param device_path: The device path of a module, which
        generally contains tty/tty* in its name.
        :returns: The matching port, or an empty port dataclass
        """
        for s in self.sorted_ports:
            vertex = self._dfs.graph.get_vertex(s)
            port = vertex.vertex
            if port.device_path.find(device_path):
                return port
        return USBPort(
            name="", sub_names=[], hub=None, port_number=None, device_path=device_path
        )

    def sort_ports(self) -> None:
        """
        Sort ports.

        Check the cached bus read vs the new bus read. Update
        graph and sorted ports accordingly.
        :param device_path: The device path of a module, which
        generally contains tty/tty* in its name.
        :returns: The matching port, or an empty port dataclass
        """
        updated_bus = self.read_usb_bus()
        remove_difference = set(self.usb_dev) - set(updated_bus)
        add_difference = set(updated_bus) - set(self.usb_dev)

        if remove_difference or add_difference:
            for d in remove_difference:
                self._dfs.graph.remove_vertex(d)
            for d in add_difference:
                self._dfs.graph.add_vertex(d)
            self.sorted_ports = self._dfs.dfs()
            self.usb_dev = updated_bus

    def match_virtual_ports(
        self, virtual_ports: List[ModuleAtPort]
    ) -> List[ModuleAtPort]:
        """
        Match Virtual Ports

        Given a list of virtual module ports, look up the
        symlink to find the device path and link that to
        the physical usb port information.
        The virtual port path looks something like:
        dev/ot_module_[MODULE NAME]
        :param virtual_ports: A list of ModuleAtPort objects
        that hold the name and virtual port of each module
        connected to the robot.

        :returns: The updated list of ModuleAtPort
        dataclasses with the physical usb port
        information updated.
        """
        self.sort_ports()
        sorted_virtual_ports = []
        for p in self.usb_dev:
            for vp in virtual_ports:
                serial_port = self.read_symlink(vp.port)
                if serial_port in p.device_path:
                    vp.usb_port = p
                    sorted_virtual_ports.append(vp)
                    break
        return sorted_virtual_ports or virtual_ports



"""Main ProtocolEngine factory."""
from opentrons.hardware_control.api import API as HardwareAPI

from .protocol_engine import ProtocolEngine
from .state import create_state_store
from .execution import create_queue_worker


async def create_protocol_engine(hardware_api: HardwareAPI) -> ProtocolEngine:
    """Create a ProtocolEngine instance.

    Arguments:
        hardware_api: Hardware control API to pass down to dependencies.
    """
    state_store = await create_state_store()

    queue_worker = create_queue_worker(
        hardware_api=hardware_api,
        state_store=state_store,
    )

    return ProtocolEngine(
        state_store=state_store,
        queue_worker=queue_worker,
        hardware_api=hardware_api,
    )



"""
hardware_control: The sole authority for controlling the hardware of an OT2.

The hardware_control module presents a unified api for the lowest level of
hardware command that takes into account the robot as a whole. For instance,
it presents an API for moving a specific pipette mount (not a specific motor
or axis)  to a deck-absolute point (not a Smoothie-coordinate point).

This module is not for use outside the opentrons api module. Higher-level
functions are available elsewhere.
"""

from .adapters import SynchronousAdapter
from .api import API
from .pause_manager import PauseManager
from .controller import Controller
from .simulator import Simulator
from .pipette import Pipette
from .types import (
    HardwareAPILike,
    CriticalPoint,
    NoTipAttachedError,
    TipAttachedError,
    ExecutionState,
    ExecutionCancelledError,
)
from .constants import DROP_TIP_RELEASE_DISTANCE
from .thread_manager import ThreadManager
from .execution_manager import ExecutionManager
from .threaded_async_lock import ThreadedAsyncLock, ThreadedAsyncForbidden

__all__ = [
    "API",
    "Controller",
    "Simulator",
    "Pipette",
    "PauseManager",
    "SynchronousAdapter",
    "HardwareAPILike",
    "CriticalPoint",
    "NoTipAttachedError",
    "TipAttachedError",
    "DROP_TIP_RELEASE_DISTANCE",
    "ThreadManager",
    "ExecutionManager",
    "ExecutionState",
    "ExecutionCancelledError",
    "ThreadedAsyncLock",
    "ThreadedAsyncForbidden",
]



from __future__ import annotations

from typing import List, Optional


class CommandBuilder:
    """Class used to build GCODE commands."""

    def __init__(self, terminator: str) -> None:
        """
        Construct a command builder.

        Args:
            terminator: The command terminator.
        """
        self._terminator = terminator
        self._elements: List[str] = []

    def add_float(
        self, prefix: str, value: float, precision: Optional[int]
    ) -> CommandBuilder:
        """
        Add a float value.

        Args:
            prefix: The value prefix.
            value: The float value.
            precision: Rounding precision. If None, there will be no rounding.

        Returns: self
        """
        value = round(value, precision) if precision is not None else value
        return self.add_element(f"{prefix}{value}")

    def add_int(self, prefix: str, value: int) -> CommandBuilder:
        """
        Add an integer value.

        Args:
            prefix: The value prefix
            value: The integer value

        Returns: self
        """
        return self.add_element(f"{prefix}{value}")

    def add_gcode(self, gcode: str) -> CommandBuilder:
        """
        Add a GCODE.

        Args:
            gcode: The gcode

        Returns: self
        """
        return self.add_element(gcode)

    def add_builder(self, builder: CommandBuilder) -> CommandBuilder:
        """
        Add all elements from builder

        Args:
            builder: a command builder

        Returns: self
        """
        self._elements += builder._elements
        return self

    def add_element(self, element: str) -> CommandBuilder:
        """
        Add an element to the command builder

        Args:
            element: an element as a string

        Returns: self
        """
        self._elements.append(element)
        return self

    def build(self) -> str:
        """
        Build command as a string.

        Returns: string
        """
        return " ".join(self._elements + [self._terminator])

    def __str__(self) -> str:
        return self.build()

    def __iter__(self):
        return iter(self._elements)

    def __bool__(self) -> bool:
        return len(self._elements) != 0

    def __eq__(self, other) -> bool:
        if isinstance(other, CommandBuilder):
            return self.build() == other.build()
        return False



import typing

from opentrons import types
from opentrons.hardware_control import NoTipAttachedError, TipAttachedError
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.types import HardwareAction
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.util import FlowRates, PlungerSpeeds, Clearances
from opentrons.protocols.context.paired_instrument import AbstractPairedInstrument
from opentrons.protocols.context.simulator.paired_instrument import (
    PairedInstrumentSimulation,
)
from opentrons.protocols.geometry import planning
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.context.protocol import AbstractProtocol
from opentrons.protocols.context.well import WellImplementation


class InstrumentContextSimulation(AbstractInstrument):
    """A simulation of an instrument context."""

    def __init__(
        self,
        protocol_interface: AbstractProtocol,
        pipette_dict: PipetteDict,
        mount: types.Mount,
        instrument_name: str,
        default_speed: float = 400.0,
        api_version: typing.Optional[APIVersion] = None,
    ):
        """Constructor."""
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._pipette_dict = pipette_dict
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._api_version = api_version or MAX_SUPPORTED_VERSION
        self._flow_rate = FlowRates(self)
        self._flow_rate.set_defaults(api_level=self._api_version)
        self._plunger_speeds = PlungerSpeeds(self)
        # Cache the maximum instrument height
        self._instrument_max_height = (
            protocol_interface.get_hardware().hardware.get_instrument_max_height(
                self._mount
            )
        )

    def get_default_speed(self) -> float:
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        self._default_speed = speed

    def aspirate(self, volume: float, rate: float) -> None:
        self._raise_if_no_tip(HardwareAction.ASPIRATE.name)
        new_volume = self.get_current_volume() + volume
        assert (
            new_volume <= self._pipette_dict["working_volume"]
        ), "Cannot aspirate more than pipette max volume"
        self._pipette_dict["current_volume"] = new_volume

    def dispense(self, volume: float, rate: float) -> None:
        self._raise_if_no_tip(HardwareAction.DISPENSE.name)
        self._pipette_dict["current_volume"] -= volume

    def blow_out(self) -> None:
        self._raise_if_no_tip(HardwareAction.BLOWOUT.name)
        self._pipette_dict["current_volume"] = 0
        self._pipette_dict["ready_to_aspirate"] = False

    def touch_tip(
        self, location: WellImplementation, radius: float, v_offset: float, speed: float
    ) -> None:
        pass

    def pick_up_tip(
        self,
        well: WellImplementation,
        tip_length: float,
        presses: typing.Optional[int],
        increment: typing.Optional[float],
    ) -> None:
        self._raise_if_tip("pick up tip")
        self._pipette_dict["has_tip"] = True
        self._pipette_dict["current_volume"] = 0

    def drop_tip(self, home_after: bool) -> None:
        self._raise_if_no_tip(HardwareAction.DROPTIP.name)
        self._pipette_dict["has_tip"] = False

    def home(self) -> None:
        self._protocol_interface.set_last_location(None)

    def home_plunger(self) -> None:
        pass

    def delay(self) -> None:
        pass

    def move_to(
        self,
        location: types.Location,
        force_direct: bool,
        minimum_z_height: typing.Optional[float],
        speed: typing.Optional[float],
    ) -> None:
        """Simulation of only the motion planning portion of move_to."""
        last_location = self._protocol_interface.get_last_location()
        if last_location:
            from_loc = last_location
        else:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        # We just want to catch planning errors.
        planning.plan_moves(
            from_loc,
            location,
            self._protocol_interface.get_deck(),
            self._instrument_max_height,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
        )

        self._protocol_interface.set_last_location(location)

    def get_mount(self) -> types.Mount:
        return self._mount

    def get_instrument_name(self) -> str:
        return self._instrument_name

    def get_pipette_name(self) -> str:
        return self._pipette_dict["name"]

    def get_model(self) -> str:
        return self._pipette_dict["model"]

    def get_min_volume(self) -> float:
        return self._pipette_dict["min_volume"]

    def get_max_volume(self) -> float:
        return self._pipette_dict["max_volume"]

    def get_current_volume(self) -> float:
        return self._pipette_dict["current_volume"]

    def get_available_volume(self) -> float:
        return self._pipette_dict["available_volume"]

    def get_pipette(self) -> PipetteDict:
        return self._pipette_dict

    def get_channels(self) -> int:
        return self._pipette_dict["channels"]

    def has_tip(self) -> bool:
        return self._pipette_dict["has_tip"]

    def is_ready_to_aspirate(self) -> bool:
        return self._pipette_dict["ready_to_aspirate"]

    def prepare_for_aspirate(self) -> None:
        self._raise_if_no_tip(HardwareAction.PREPARE_ASPIRATE.name)

    def get_return_height(self) -> float:
        return self._pipette_dict["return_tip_height"]

    def get_well_bottom_clearance(self) -> Clearances:
        return Clearances(default_aspirate=1, default_dispense=1)

    def get_speed(self) -> PlungerSpeeds:
        return self._plunger_speeds

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rate

    def set_flow_rate(
        self,
        aspirate: typing.Optional[float] = None,
        dispense: typing.Optional[float] = None,
        blow_out: typing.Optional[float] = None,
    ) -> None:
        if aspirate is not None:
            self._pipette_dict["aspirate_flow_rate"] = aspirate
        if dispense is not None:
            self._pipette_dict["dispense_flow_rate"] = dispense
        if blow_out is not None:
            self._pipette_dict["blow_out_flow_rate"] = blow_out

    def set_pipette_speed(
        self,
        aspirate: typing.Optional[float] = None,
        dispense: typing.Optional[float] = None,
        blow_out: typing.Optional[float] = None,
    ) -> None:
        if aspirate is not None:
            self._pipette_dict["aspirate_speed"] = aspirate
        if dispense is not None:
            self._pipette_dict["dispense_speed"] = dispense
        if blow_out is not None:
            self._pipette_dict["blow_out_speed"] = blow_out

    def _raise_if_no_tip(self, action: str) -> None:
        """Raise NoTipAttachedError if no tip."""
        if not self.has_tip():
            raise NoTipAttachedError(f"Cannot perform {action} without a tip attached")

    def _raise_if_tip(self, action: str) -> None:
        """Raise TipAttachedError if tip."""
        if self.has_tip():
            raise TipAttachedError(f"Cannot {action} with a tip attached")

    def pair_with(
        self, other_instrument: AbstractInstrument
    ) -> AbstractPairedInstrument:
        return PairedInstrumentSimulation(
            primary_instrument=self,
            secondary_instrument=other_instrument,
            ctx=self._protocol_interface,
        )



"""Translation of JSON protocol commands into ProtocolEngine commands."""
from typing import Dict, List, cast

from opentrons.types import DeckSlotName, MountType
from opentrons.protocols import models
from opentrons.protocol_engine import (
    commands as pe_commands,
    DeckSlotLocation,
    PipetteName,
    WellLocation,
    WellOrigin,
)


class CommandTranslatorError(Exception):
    """An error raised to indicate an internal programmer error."""

    pass


class JsonCommandTranslator:
    """Class that translates commands from PD/JSON to ProtocolEngine."""

    def translate(
        self,
        protocol: models.JsonProtocol,
    ) -> List[pe_commands.CommandRequest]:
        """Return all Protocol Engine commands required to run the given protocol."""
        result: List[pe_commands.CommandRequest] = []

        for pipette_id, pipette in protocol.pipettes.items():
            result.append(self._translate_load_pipette(pipette_id, pipette))

        for definition_id, definition in protocol.labwareDefinitions.items():
            result.append(self._translate_add_labware_definition(definition))

        for labware_id, labware in protocol.labware.items():
            result.append(
                self._translate_load_labware(
                    labware_id, labware, protocol.labwareDefinitions
                )
            )

        for command in protocol.commands:
            result.append(self._translate_command(command))

        return result

    def _translate_command(
        self,
        command: models.json_protocol.AllCommands,
    ) -> pe_commands.CommandRequest:
        try:
            h = self._COMMAND_TO_NAME[command.command]
            return cast(pe_commands.CommandRequest, getattr(self, h)(command))
        except KeyError:
            raise CommandTranslatorError(f"'{command.command}' is not recognized.")
        except AttributeError:
            raise CommandTranslatorError(
                f"Cannot find handler for '{command.command}'."
            )

    def _translate_add_labware_definition(
        self, labware_definition: models.LabwareDefinition
    ) -> pe_commands.AddLabwareDefinitionRequest:
        return pe_commands.AddLabwareDefinitionRequest(
            data=pe_commands.AddLabwareDefinitionData(definition=labware_definition)
        )

    def _translate_load_labware(
        self,
        labware_id: str,
        labware: models.json_protocol.Labware,
        labware_definitions: Dict[str, models.LabwareDefinition],
    ) -> pe_commands.LoadLabwareRequest:
        """Translate a JSON labware data into a LoadLabware command.

        Args:
            labware_id: The ID that the JSON protocol's commands will use to refer to
                this labware placement.
            labware: The JSON protocol's details about this labware placement, including
                which deck slot it should go in, and a pointer to a labware definition.
            labware_definitions: The JSON protocol's collection of labware definitions.
        """
        definition = labware_definitions[labware.definitionId]
        return pe_commands.LoadLabwareRequest(
            data=pe_commands.LoadLabwareData(
                location=DeckSlotLocation(
                    slot=DeckSlotName.from_primitive(labware.slot)
                ),
                loadName=definition.parameters.loadName,
                namespace=definition.namespace,
                version=definition.version,
                labwareId=labware_id,
            )
        )

    def _translate_load_pipette(
        self,
        pipette_id: str,
        pipette: models.json_protocol.Pipettes,
    ) -> pe_commands.LoadPipetteRequest:
        return pe_commands.LoadPipetteRequest(
            data=pe_commands.LoadPipetteData(
                pipetteName=PipetteName(pipette.name),
                mount=MountType(pipette.mount),
                pipetteId=pipette_id,
            )
        )

    def _aspirate(
        self,
        command: models.json_protocol.LiquidCommand,
    ) -> pe_commands.AspirateRequest:
        return pe_commands.AspirateRequest(
            data=pe_commands.AspirateData(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
                volume=command.params.volume,
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=(0, 0, command.params.offsetFromBottomMm),
                ),
            )
        )

    def _dispense(
        self,
        command: models.json_protocol.LiquidCommand,
    ) -> pe_commands.DispenseRequest:
        return pe_commands.DispenseRequest(
            data=pe_commands.DispenseData(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
                volume=command.params.volume,
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=(0, 0, command.params.offsetFromBottomMm),
                ),
            )
        )

    def _air_gap(self, command: models.json_protocol.LiquidCommand) -> None:
        raise NotImplementedError()

    def _blowout(self, command: models.json_protocol.BlowoutCommand) -> None:
        raise NotImplementedError()

    def _touch_tip(self, command: models.json_protocol.TouchTipCommand) -> None:
        raise NotImplementedError()

    def _pick_up(
        self,
        command: models.json_protocol.PickUpDropTipCommand,
    ) -> pe_commands.PickUpTipRequest:
        return pe_commands.PickUpTipRequest(
            data=pe_commands.PickUpTipData(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
            )
        )

    def _drop_tip(
        self,
        command: models.json_protocol.PickUpDropTipCommand,
    ) -> pe_commands.DropTipRequest:
        return pe_commands.DropTipRequest(
            data=pe_commands.DropTipData(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
            )
        )

    def _move_to_slot(
        self,
        command: models.json_protocol.MoveToSlotCommand,
    ) -> None:
        raise NotImplementedError()

    def _delay(
        self,
        command: models.json_protocol.DelayCommand,
    ) -> pe_commands.PauseRequest:
        if command.params.wait is not True:
            raise NotImplementedError("Delay translation not yet implemented.")

        data = pe_commands.PauseData(message=command.params.message)
        return pe_commands.PauseRequest(data=data)

    def _magnetic_module_engage(
        self,
        command: models.json_protocol.MagneticModuleEngageCommand,
    ) -> None:
        raise NotImplementedError()

    def _magnetic_module_disengage(
        self,
        command: models.json_protocol.MagneticModuleDisengageCommand,
    ) -> None:
        raise NotImplementedError()

    def _temperature_module_set_target(
        self,
        command: models.json_protocol.TemperatureModuleSetTargetCommand,
    ) -> None:
        raise NotImplementedError()

    def _temperature_module_await_temperature(
        self,
        command: models.json_protocol.TemperatureModuleAwaitTemperatureCommand,
    ) -> None:
        raise NotImplementedError()

    def _temperature_module_deactivate(
        self,
        command: models.json_protocol.TemperatureModuleDeactivateCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_set_target_block_temperature(
        self,
        command: models.json_protocol.ThermocyclerSetTargetBlockTemperatureCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_set_target_lid_temperature(
        self, command: models.json_protocol.ThermocyclerSetTargetLidTemperatureCommand
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_await_block_temperature(
        self, command: models.json_protocol.ThermocyclerAwaitBlockTemperatureCommand
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_await_lid_temperature(
        self,
        command: models.json_protocol.ThermocyclerAwaitLidTemperatureCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_deactivate_block(
        self,
        command: models.json_protocol.ThermocyclerDeactivateBlockCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_deactivate_lid(
        self,
        command: models.json_protocol.ThermocyclerDeactivateLidCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_open_lid(
        self,
        command: models.json_protocol.ThermocyclerOpenLidCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_close_lid(
        self,
        command: models.json_protocol.ThermocyclerCloseLidCommand,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_run_profile(
        self,
        command: models.json_protocol.ThermocyclerRunProfile,
    ) -> None:
        raise NotImplementedError()

    def _thermocycler_await_profile_complete(
        self,
        command: models.json_protocol.ThermocyclerAwaitProfileCompleteCommand,
    ) -> None:
        raise NotImplementedError()

    def _move_to_well(
        self,
        command: models.json_protocol.MoveToWellCommand,
    ) -> None:
        raise NotImplementedError()

    _COMMAND_TO_NAME: Dict[str, str] = {
        models.json_protocol.CommandMoveToWell: _move_to_well.__name__,
        models.json_protocol.CommandThermocyclerAwaitProfile: _thermocycler_await_profile_complete.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerRunProfile: _thermocycler_run_profile.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerCloseLid: _thermocycler_close_lid.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerOpenLid: _thermocycler_open_lid.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerDeactivateLid: _thermocycler_deactivate_lid.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerDeactivateBlock: _thermocycler_deactivate_block.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerSetTargetLid: _thermocycler_set_target_lid_temperature.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerAwaitBlockTemperature: _thermocycler_await_block_temperature.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerAwaitLidTemperature: _thermocycler_await_lid_temperature.__name__,  # noqa: E501
        models.json_protocol.CommandThermocyclerSetTargetBlock: _thermocycler_set_target_block_temperature.__name__,  # noqa: E501
        models.json_protocol.CommandTemperatureModuleDeactivate: _temperature_module_deactivate.__name__,  # noqa: E501
        models.json_protocol.CommandTemperatureModuleAwait: _temperature_module_await_temperature.__name__,  # noqa: E501
        models.json_protocol.CommandTemperatureModuleSetTarget: _temperature_module_set_target.__name__,  # noqa: E501
        models.json_protocol.CommandMagneticModuleDisengage: _magnetic_module_disengage.__name__,  # noqa: E501
        models.json_protocol.CommandMagneticModuleEngage: _magnetic_module_engage.__name__,  # noqa: E501
        models.json_protocol.CommandDelay: _delay.__name__,
        models.json_protocol.CommandMoveToSlot: _move_to_slot.__name__,
        models.json_protocol.CommandDropTip: _drop_tip.__name__,
        models.json_protocol.CommandPickUpTip: _pick_up.__name__,
        models.json_protocol.CommandTouchTip: _touch_tip.__name__,
        models.json_protocol.CommandBlowout: _blowout.__name__,
        models.json_protocol.CommandAirGap: _air_gap.__name__,
        models.json_protocol.CommandDispense: _dispense.__name__,
        models.json_protocol.CommandAspirate: _aspirate.__name__,
    }



"""StateStore and dependency factory."""
from ..resources import DeckDataProvider
from .state import StateStore


async def create_state_store() -> StateStore:
    """Create a ready-to-use StateStore instance."""
    deck_data = DeckDataProvider()

    # TODO(mc, 2020-11-18): check short trash FF
    deck_definition = await deck_data.get_deck_definition()
    deck_fixed_labware = await deck_data.get_deck_fixed_labware(deck_definition)

    return StateStore(
        deck_definition=deck_definition,
        deck_fixed_labware=deck_fixed_labware,
    )



"""Asynchronous task queue to accomplish a protocol run."""
import asyncio
from enum import Enum
from typing import Any, Awaitable, Callable, Dict, NamedTuple, Optional, Tuple


class TaskQueuePhase(str, Enum):
    """Phase in which to run a given task."""

    RUN = "run"
    CLEANUP = "cleanup"


class TaskQueueEntry(NamedTuple):
    """An entry in the task queue."""

    func: Callable[..., Awaitable[Any]]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


class TaskQueue:
    """A queue of async tasks to run.

    Once started, a TaskQueue may not be re-used.
    """

    def __init__(self) -> None:
        """Initialize the TaskQueue."""
        self._run_entry: Optional[TaskQueueEntry] = None
        self._cleanup_entry: Optional[TaskQueueEntry] = None
        self._run_task: Optional["asyncio.Task[None]"] = None
        self._run_started_event: asyncio.Event = asyncio.Event()

    def add(
        self,
        phase: TaskQueuePhase,
        func: Callable[..., Awaitable[Any]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Add a task to the queue.

        Two phases are available: TaskQueuePhase.RUN and TaskQueuePhase.CLEANUP.
        Each phase may contain no tasks or one task. The RUN task will be run first,
        if present, and the CLEANUP task will run second. The CLEANUP task will run
        regardless of the success or failure of the RUN task.
        """
        entry = TaskQueueEntry(func=func, args=args, kwargs=kwargs)

        if phase == TaskQueuePhase.RUN:
            self._run_entry = entry
        else:
            self._cleanup_entry = entry

    def is_started(self) -> bool:
        """Get whether the task queue has started running."""
        return self._run_task is not None

    def start(self) -> None:
        """Start running tasks in the queue."""
        self._run_task = asyncio.create_task(self._run())
        self._run_started_event.set()

    async def join(self) -> None:
        """Wait for the background run task to complete, propagating errors."""
        await self._run_started_event.wait()

        if self._run_task:
            await self._run_task

    async def _run(self) -> None:
        run_entry = self._run_entry
        cleanup_entry = self._cleanup_entry

        try:
            await self._run_task_entry(run_entry)
        finally:
            await self._run_task_entry(cleanup_entry)

    @staticmethod
    async def _run_task_entry(entry: Optional[TaskQueueEntry]) -> None:
        if entry:
            await entry.func(*entry.args, **entry.kwargs)



from .protocol_context import ProtocolContext
from .instrument_context import InstrumentContext
from .paired_instrument_context import PairedInstrumentContext
from .module_contexts import (
    ModuleContext,
    ThermocyclerContext,
    MagneticModuleContext,
    TemperatureModuleContext,
)


__all__ = [
    "ProtocolContext",
    "InstrumentContext",
    "ModuleContext",
    "PairedInstrumentContext",
    "ThermocyclerContext",
    "MagneticModuleContext",
    "TemperatureModuleContext",
]



from itertools import dropwhile, takewhile
from typing import Optional, Sequence

from opentrons.protocols.context.well import WellImplementation


Wells = Sequence[WellImplementation]
WellColumns = Sequence[Wells]


class TipTracker:
    def __init__(self, columns: WellColumns):
        self._columns = columns

    def next_tip(
        self, num_tips: int = 1, starting_tip: Optional[WellImplementation] = None
    ) -> Optional[WellImplementation]:
        """
        Find the next valid well for pick-up.

        Determines the next valid start tip from which to retrieve the
        specified number of tips. There must be at least `num_tips` sequential
        wells for which all wells have tips, in the same column.

        :param num_tips: target number of sequential tips in the same column
        :type num_tips: int
        :param starting_tip: The :py:class:`.Well` from which to start search.
                for an available tip.
        :type starting_tip: :py:class:`.Well`
        :return: the :py:class:`.Well` meeting the target criteria, or None
        """
        columns: WellColumns = self._columns
        if starting_tip:
            # Remove columns preceding the one with the pipette's starting tip
            drop_undefined_columns = list(
                dropwhile(lambda x: starting_tip not in x, columns)
            )
            # Remove tips preceding the starting tip in the first column
            drop_undefined_columns[0] = list(
                dropwhile(lambda w: starting_tip is not w, drop_undefined_columns[0])
            )
            columns = drop_undefined_columns

        drop_leading_empties = [
            list(dropwhile(lambda x: not x.has_tip(), column)) for column in columns
        ]
        drop_at_first_gap = [
            list(takewhile(lambda x: x.has_tip(), column))
            for column in drop_leading_empties
        ]
        long_enough = [
            column for column in drop_at_first_gap if len(column) >= num_tips
        ]

        try:
            first_long_enough = long_enough[0]
            result: Optional[WellImplementation] = first_long_enough[0]
        except IndexError:
            result = None

        return result

    def use_tips(
        self,
        start_well: WellImplementation,
        num_channels: int = 1,
        fail_if_full: bool = False,
    ):
        """
        Removes tips from the tip tracker.

        This method should be called when a tip is picked up. Generally, it
        will be called with `num_channels=1` or `num_channels=8` for single-
        and multi-channel respectively. If picking up with more than one
        channel, this method will automatically determine which tips are used
        based on the start well, the number of channels, and the geometry of
        the tiprack.

        :param start_well: The :py:class:`.Well` from which to pick up a tip.
                           For a single-channel pipette, this is the well to
                           send the pipette to. For a multi-channel pipette,
                           this is the well to send the back-most nozzle of the
                           pipette to.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int
        :param fail_if_full: for backwards compatibility
        """
        # Select the column of the labware that contains the target well
        target_column = [col for col in self._columns if start_well in col][0]

        well_idx = target_column.index(start_well)
        # Number of tips to pick up is the lesser of (1) the number of tips
        # from the starting well to the end of the column, and (2) the number
        # of channels of the pipette (so a 4-channel pipette would pick up a
        # max of 4 tips, and picking up from the 2nd-to-bottom well in a
        # column would get a maximum of 2 tips)
        num_tips = min(len(target_column) - well_idx, num_channels)
        target_wells = target_column[well_idx : well_idx + num_tips]

        # In API version 2.2, we no longer reset the tip tracker when a tip
        # is dropped back into a tiprack well. This fixes a behavior where
        # subsequent transfers would reuse the dirty tip. However, sometimes
        # the user explicitly wants to use a dirty tip, and this check would
        # raise an exception if they tried to do so.
        # An extension of work here is to have separate tip trackers for
        # dirty tips and non-present tips; but until then, we can avoid the
        # exception.
        if fail_if_full:
            assert all(
                well.has_tip() for well in target_wells
            ), "{} is out of tips".format(str(self))

        for well in target_wells:
            well.set_has_tip(False)

    def previous_tip(self, num_tips: int = 1) -> Optional[WellImplementation]:
        """
        Find the best well to drop a tip in.

        This is the well from which the last tip was picked up, if there's
        room. It can be used to return tips to the tip tracker.

        :param num_tips: target number of tips to return, sequential in a
                         column
        :type num_tips: int
        :return: The :py:class:`.Well` meeting the target criteria, or ``None``
        """
        columns = self._columns
        drop_leading_filled = [
            list(dropwhile(lambda x: x.has_tip(), column)) for column in columns
        ]
        drop_at_first_gap = [
            list(takewhile(lambda x: not x.has_tip(), column))
            for column in drop_leading_filled
        ]
        long_enough = [
            column for column in drop_at_first_gap if len(column) >= num_tips
        ]
        try:
            return long_enough[0][0]
        except IndexError:
            return None

    def return_tips(self, start_well: WellImplementation, num_channels: int = 1):
        """
        Re-adds tips to the tip tracker

        This method should be called when a tip is dropped in a tiprack. It
        should be called with ``num_channels=1`` or ``num_channels=8`` for
        single- and multi-channel respectively. If returning more than one
        channel, this method will automatically determine which tips are
        returned based on the start well, the number of channels,
        and the tiprack geometry.

        Note that unlike :py:meth:`use_tips`, calling this method in a way
        that would drop tips into wells with tips in them will raise an
        exception; this should only be called on a valid return of
        :py:meth:`previous_tip`.

        :param start_well: The :py:class:`.Well` into which to return a tip.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int
        """
        # Select the column that contains the target_well
        target_column = [col for col in self._columns if start_well in col][0]
        well_idx = target_column.index(start_well)
        end_idx = min(well_idx + num_channels, len(target_column))
        drop_targets = target_column[well_idx:end_idx]
        for well in drop_targets:
            if well.has_tip():
                raise AssertionError(f"Well {repr(well)} has a tip")
        for well in drop_targets:
            well.set_has_tip(True)



"""Union types of concrete command definitions."""
from typing import Union

from .add_labware_definition import (
    AddLabwareDefinition,
    AddLabwareDefinitionRequest,
    AddLabwareDefinitionResult,
    AddLabwareDefinitionCommandType,
)

from .aspirate import Aspirate, AspirateRequest, AspirateResult, AspirateCommandType

from .dispense import Dispense, DispenseRequest, DispenseResult, DispenseCommandType

from .drop_tip import DropTip, DropTipRequest, DropTipResult, DropTipCommandType

from .load_labware import (
    LoadLabware,
    LoadLabwareRequest,
    LoadLabwareResult,
    LoadLabwareCommandType,
)

from .load_pipette import (
    LoadPipette,
    LoadPipetteRequest,
    LoadPipetteResult,
    LoadPipetteCommandType,
)

from .move_to_well import (
    MoveToWell,
    MoveToWellRequest,
    MoveToWellResult,
    MoveToWellCommandType,
)

from .pick_up_tip import (
    PickUpTip,
    PickUpTipRequest,
    PickUpTipResult,
    PickUpTipCommandType,
)

from .pause import (
    Pause,
    PauseRequest,
    PauseResult,
    PauseCommandType,
)

Command = Union[
    AddLabwareDefinition,
    Aspirate,
    Dispense,
    DropTip,
    LoadLabware,
    LoadPipette,
    MoveToWell,
    PickUpTip,
    Pause,
]

CommandType = Union[
    AddLabwareDefinitionCommandType,
    AspirateCommandType,
    DispenseCommandType,
    DropTipCommandType,
    LoadLabwareCommandType,
    LoadPipetteCommandType,
    MoveToWellCommandType,
    PickUpTipCommandType,
    PauseCommandType,
]

CommandRequest = Union[
    AddLabwareDefinitionRequest,
    AspirateRequest,
    DispenseRequest,
    DropTipRequest,
    LoadLabwareRequest,
    LoadPipetteRequest,
    MoveToWellRequest,
    PickUpTipRequest,
    PauseRequest,
]

CommandResult = Union[
    AddLabwareDefinitionResult,
    AspirateResult,
    DispenseResult,
    DropTipResult,
    LoadLabwareResult,
    LoadPipetteResult,
    MoveToWellResult,
    PickUpTipResult,
    PauseResult,
]



import asyncio
import logging
from typing import Mapping, Optional
from opentrons.drivers.mag_deck import (
    SimulatingDriver,
    MagDeckDriver,
    AbstractMagDeckDriver,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from . import update, mod_abc, types

log = logging.getLogger(__name__)

MAX_ENGAGE_HEIGHT = {
    # mm from home position
    "magneticModuleV1": 45,
    "magneticModuleV2": 25,
}
OFFSET_TO_LABWARE_BOTTOM = {"magneticModuleV1": 5, "magneticModuleV2": 2.5}


class MagDeck(mod_abc.AbstractModule):

    FIRST_GEN2_REVISION = 20

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        interrupt_callback: types.InterruptCallback = None,
        simulating=False,
        loop: asyncio.AbstractEventLoop = None,
        sim_model: str = None,
        **kwargs,
    ):
        """Factory function."""
        driver: AbstractMagDeckDriver
        if not simulating:
            driver = await MagDeckDriver.create(port=port, loop=loop)
        else:
            driver = SimulatingDriver(sim_model=sim_model)

        mod = cls(
            port=port,
            usb_port=usb_port,
            loop=loop,
            execution_manager=execution_manager,
            device_info=await driver.get_device_info(),
            driver=driver,
        )
        return mod

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractMagDeckDriver,
        device_info: Mapping[str, str],
        loop: asyncio.AbstractEventLoop = None,
    ) -> None:
        """Constructor"""
        super().__init__(
            port=port, usb_port=usb_port, loop=loop, execution_manager=execution_manager
        )
        self._device_info = device_info
        self._driver = driver
        self._current_height = 0.0

    @classmethod
    def name(cls) -> str:
        """Get the module name."""
        return "magdeck"

    def model(self) -> str:
        """Get the model."""
        return self._model_from_revision(self._device_info.get("model"))

    @classmethod
    def bootloader(cls) -> types.UploadFunction:
        """Get the bootloating method."""
        return update.upload_via_avrdude

    async def calibrate(self):
        """Calibration involves probing for top plate to get the plate height."""
        await self.wait_for_is_running()
        await self._driver.probe_plate()
        # return if successful or not?

    async def engage(self, height: float):
        """Move the magnet to a specific height, in mm from home position."""
        await self.wait_for_is_running()
        if height > MAX_ENGAGE_HEIGHT[self.model()] or height < 0:
            raise ValueError(
                f"Invalid engage height for {self.model()}: {height} mm. "
                f"Must be 0 - {MAX_ENGAGE_HEIGHT[self.model()]} mm"
            )
        await self._driver.move(height)
        self._current_height = await self._driver.get_mag_position()

    async def deactivate(self):
        """Home the magnet."""
        await self.wait_for_is_running()
        await self._driver.home()
        await self.engage(0.0)

    @property
    def current_height(self) -> float:
        """Get the current height."""
        return self._current_height

    @property
    def device_info(self) -> Mapping[str, str]:
        """

        Returns: a dict
            { 'serial': 'abc123', 'model': '8675309', 'version': '9001' }

        """
        return self._device_info

    @property
    def status(self) -> str:
        if self.current_height > 0:
            return "engaged"
        else:
            return "disengaged"

    @property
    def engaged(self) -> bool:
        if self.current_height > 0:
            return True
        else:
            return False

    @property
    def live_data(self) -> types.LiveData:
        return {
            "status": self.status,
            "data": {"engaged": self.engaged, "height": self.current_height},
        }

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    # Internal Methods

    async def prep_for_update(self) -> str:
        await self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        if not revision or "v" not in revision:
            log.error(f"bad revision: {revision}")
            return "magneticModuleV1"
        try:
            revision_num = float(revision.split("v")[-1])
        except (ValueError, TypeError):
            log.exception("bad revision: {revision}")
            return "magneticModuleV1"
        if revision_num < MagDeck.FIRST_GEN2_REVISION:
            return "magneticModuleV1"
        else:
            return "magneticModuleV2"



import json
import os
import sys
from functools import partial
from pathlib import Path
import logging
import asyncio
import re
from typing import Any, List, Tuple, cast

from opentrons.drivers.serial_communication import get_ports_by_name
from opentrons.hardware_control import API, ThreadManager
from opentrons.config import (
    feature_flags as ff,
    name,
    robot_configs,
    IS_ROBOT,
    ROBOT_FIRMWARE_DIR,
)
from opentrons.util import logging_config
from opentrons.protocols.types import ApiDeprecationError
from opentrons.protocols.api_support.types import APIVersion

version = sys.version_info[0:2]
if version < (3, 7):
    raise RuntimeError(
        "opentrons requires Python 3.7 or above, this is {0}.{1}".format(
            version[0], version[1]
        )
    )

HERE = os.path.abspath(os.path.dirname(__file__))

try:
    with open(os.path.join(HERE, "package.json")) as pkg:
        package_json = json.load(pkg)
        __version__ = package_json.get("version")
except (FileNotFoundError, OSError):
    __version__ = "unknown"

from opentrons import config  # noqa: E402

LEGACY_MODULES = ["robot", "reset", "instruments", "containers", "labware", "modules"]

__all__ = ["version", "HERE", "config"]


def __getattr__(attrname):
    """
    Prevent import of legacy modules from global to officially
    deprecate Python API Version 1.0.
    """
    if attrname in LEGACY_MODULES:
        raise ApiDeprecationError(APIVersion(1, 0))
    raise AttributeError(attrname)


def __dir__():
    return sorted(__all__ + LEGACY_MODULES)


log = logging.getLogger(__name__)

try:
    import systemd.daemon  # type: ignore

    systemdd_notify = partial(systemd.daemon.notify, "READY=1")
except ImportError:
    log.info("Systemd couldn't be imported, not notifying")
    systemdd_notify = partial(lambda: None)


SMOOTHIE_HEX_RE = re.compile("smoothie-(.*).hex")


def _find_smoothie_file() -> Tuple[Path, str]:
    resources: List[Path] = []

    # Search for smoothie files in /usr/lib/firmware first then fall back to
    # value packed in wheel
    if IS_ROBOT:
        resources.extend(ROBOT_FIRMWARE_DIR.iterdir())  # type: ignore

    resources_path = Path(HERE) / "resources"
    resources.extend(resources_path.iterdir())

    for path in resources:
        matches = SMOOTHIE_HEX_RE.search(path.name)
        if matches:
            branch_plus_ref = matches.group(1)
            return path, branch_plus_ref
    raise OSError(f"Could not find smoothie firmware file in {resources_path}")


def _get_motor_control_serial_port() -> Any:
    port = os.environ.get("OT_SMOOTHIE_EMULATOR_URI")

    if port is None:
        smoothie_id = os.environ.get("OT_SMOOTHIE_ID", "AMA")
        # TODO(mc, 2021-08-01): raise a more informative exception than
        # IndexError if a valid serial port is not found
        port = get_ports_by_name(device_name=smoothie_id)[0]

    log.info(f"Connecting to motor controller at port {port}")
    return port


async def _create_thread_manager() -> ThreadManager:
    """Build the hardware controller wrapped in a ThreadManager.

    .. deprecated:: 4.6
        ThreadManager is on its way out.
    """
    if os.environ.get("ENABLE_VIRTUAL_SMOOTHIE"):
        log.info("Initialized robot using virtual Smoothie")
        return ThreadManager(API.build_hardware_simulator)

    thread_manager = ThreadManager(
        API.build_hardware_controller,
        threadmanager_nonblocking=True,
        port=_get_motor_control_serial_port(),
        firmware=_find_smoothie_file(),
    )

    try:
        await thread_manager.managed_thread_ready_async()
    except RuntimeError:
        log.exception("Could not build hardware controller, forcing virtual")
        return ThreadManager(API.build_hardware_simulator)

    return thread_manager


async def _create_hardware_api() -> API:
    use_hardware_simulator = os.environ.get("ENABLE_VIRTUAL_SMOOTHIE")

    if use_hardware_simulator:
        return await API.build_hardware_simulator()

    return await API.build_hardware_controller(
        port=_get_motor_control_serial_port(),
        firmware=_find_smoothie_file(),
    )


async def initialize() -> API:
    """
    Initialize the Opentrons hardware returning a hardware instance.
    """
    robot_conf = robot_configs.load()
    logging_config.log_init(robot_conf.log_level)

    log.info(f"API server version: {__version__}")
    log.info(f"Robot Name: {name()}")
    systemdd_notify()

    use_thread_manager = ff.enable_protocol_engine() is False

    if use_thread_manager:
        thread_manager = await _create_thread_manager()
        hardware = cast(API, thread_manager)
    else:
        hardware = await _create_hardware_api()

    async def _blink():
        while True:
            await hardware.set_lights(button=True)
            await asyncio.sleep(0.5)
            await hardware.set_lights(button=False)
            await asyncio.sleep(0.5)

    blink_task = asyncio.create_task(_blink())

    if not ff.disable_home_on_boot():
        log.info("Homing Z axes")
        await hardware.home_z()

    blink_task.cancel()
    await asyncio.gather(blink_task, return_exceptions=True)
    await hardware.set_lights(button=True)

    return hardware



import asyncio
import logging
from typing import Mapping, Optional

from opentrons.hardware_control.modules.types import TemperatureStatus
from opentrons.hardware_control.poller import Reader, WaitableListener, Poller
from typing_extensions import Final
from opentrons.drivers.types import Temperature
from opentrons.drivers.temp_deck import (
    SimulatingDriver,
    AbstractTempDeckDriver,
    TempDeckDriver,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.modules import update, mod_abc, types

log = logging.getLogger(__name__)

TEMP_POLL_INTERVAL_SECS = 1.0
SIM_TEMP_POLL_INTERVAL_SECS = TEMP_POLL_INTERVAL_SECS / 20.0


class TempDeck(mod_abc.AbstractModule):
    """
    Under development. API subject to change without a version bump
    """

    FIRST_GEN2_REVISION = 20

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        interrupt_callback: types.InterruptCallback = None,
        simulating: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        sim_model: str = None,
        **kwargs,
    ):
        """
        Build a TempDeck

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            interrupt_callback: Optional interrupt callback
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            **kwargs: Module specific values.
                can be 'polling_frequency' to specify the polling frequency in
                seconds

        Returns:
            Tempdeck instance
        """
        polling_frequency = kwargs.get("polling_frequency")
        driver: AbstractTempDeckDriver
        if not simulating:
            driver = await TempDeckDriver.create(port=port, loop=loop)
            polling_frequency = polling_frequency or TEMP_POLL_INTERVAL_SECS
        else:
            driver = SimulatingDriver(sim_model=sim_model)
            polling_frequency = polling_frequency or SIM_TEMP_POLL_INTERVAL_SECS

        mod = cls(
            port=port,
            usb_port=usb_port,
            execution_manager=execution_manager,
            driver=driver,
            device_info=await driver.get_device_info(),
            loop=loop,
            polling_frequency=polling_frequency,
        )
        return mod

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractTempDeckDriver,
        device_info: Mapping[str, str],
        loop: asyncio.AbstractEventLoop = None,
        polling_frequency: float = TEMP_POLL_INTERVAL_SECS,
    ) -> None:
        """Constructor"""
        super().__init__(
            port=port, usb_port=usb_port, loop=loop, execution_manager=execution_manager
        )
        self._device_info = device_info
        self._driver = driver
        self._listener = TempdeckListener(loop=loop)
        self._poller = Poller(
            reader=PollerReader(driver=self._driver),
            interval_seconds=polling_frequency,
            listener=self._listener,
            loop=loop,
        )

    async def cleanup(self) -> None:
        """Stop the poller task."""
        await self._poller.stop_and_wait()

    @classmethod
    def name(cls) -> str:
        return "tempdeck"

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    @classmethod
    def bootloader(cls) -> types.UploadFunction:
        return update.upload_via_avrdude

    async def wait_next_poll(self) -> None:
        """Wait for the next poll to complete."""
        await self._listener.wait_next_poll()

    async def set_temperature(self, celsius: float) -> None:
        """
        Set temperature in degree Celsius
        Range: 4 to 95 degree Celsius (QA tested).
        The internal temp range is -9 to 99 C, which is limited by the 2-digit
        temperature display. Any input outside of this range will be clipped
        to the nearest limit
        """
        await self.wait_for_is_running()
        await self._driver.set_temperature(celsius=celsius)
        await self.wait_next_poll()

        async def _wait():
            # Wait until we reach the target temperature.
            while self.status != TemperatureStatus.HOLDING:
                await self.wait_next_poll()

        task = self._loop.create_task(_wait())
        await self.make_cancellable(task)
        await task

    async def start_set_temperature(self, celsius) -> None:
        """
        Set temperature in degree Celsius
        Range: 4 to 95 degree Celsius (QA tested).
        The internal temp range is -9 to 99 C, which is limited by the 2-digit
        temperature display. Any input outside of this range will be clipped
        to the nearest limit
        """
        await self.wait_for_is_running()
        await self._driver.set_temperature(celsius)

    async def await_temperature(self, awaiting_temperature: float) -> None:
        """
        Await temperature in degree Celsius
        Polls temperature module's temperature until
        the specified temperature is reached
        """
        if self.is_simulated:
            return

        await self.wait_for_is_running()
        await self.wait_next_poll()

        async def _await_temperature():
            if self.status == TemperatureStatus.HEATING:
                while self.temperature < awaiting_temperature:
                    await self.wait_next_poll()
            elif self.status == TemperatureStatus.COOLING:
                while self.temperature > awaiting_temperature:
                    await self.wait_next_poll()

        t = self._loop.create_task(_await_temperature())
        await self.make_cancellable(t)
        await t

    async def deactivate(self):
        """Stop heating/cooling and turn off the fan"""
        await self.wait_for_is_running()
        await self._driver.deactivate()

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def live_data(self) -> types.LiveData:
        return {
            "status": self.status,
            "data": {"currentTemp": self.temperature, "targetTemp": self.target},
        }

    @property
    def temperature(self) -> float:
        return self._listener.state.current

    @property
    def target(self) -> Optional[float]:
        return self._listener.state.target

    @property
    def status(self) -> str:
        return self._get_status(self._listener.state).value

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    async def prep_for_update(self) -> str:
        model = self._device_info and self._device_info.get("model")
        if model in ("temp_deck_v1", "temp_deck_v1.1", "temp_deck_v2"):
            raise types.UpdateError(
                "This Temperature Module can't be updated."
                "Please contact Opentrons Support."
            )
        await self._poller.stop_and_wait()
        await self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port

    def has_available_update(self) -> bool:
        """Override of abc implementation to suppress update notifications
        for v1, v1.1, and v2 temperature modules which cannot be updated"""
        if not self._device_info:
            model = None
        else:
            model = self._device_info.get("model")
        if model in {"temp_deck_v1", "temp_deck_v1.1", "temp_deck_v2", None}:
            return False
        return super().has_available_update()

    @staticmethod
    def _get_status(temperature: Temperature) -> TemperatureStatus:
        """
        Determine the status from the temperature.

        Args:
            temperature: A Temperature instance

        Returns:
            The status
        """
        DELTA: Final = 0.7
        status = TemperatureStatus.IDLE
        if temperature.target is not None:
            diff = temperature.target - temperature.current
            if abs(diff) < DELTA:  # To avoid status fluctuation near target
                status = TemperatureStatus.HOLDING
            elif diff < 0:
                status = TemperatureStatus.COOLING
            else:
                status = TemperatureStatus.HEATING
        return status

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        if not revision or "v" not in revision:
            log.error(f"bad revision: {revision}")
            return "temperatureModuleV1"
        try:
            revision_num = float(revision.split("v")[-1])
        except (ValueError, TypeError):
            # none or corrupt
            log.exception("no revision")
            return "temperatureModuleV1"

        if revision_num < TempDeck.FIRST_GEN2_REVISION:
            return "temperatureModuleV1"
        else:
            return "temperatureModuleV2"


class PollerReader(Reader[Temperature]):
    """Polled data reader."""

    def __init__(self, driver: AbstractTempDeckDriver) -> None:
        """Constructor."""
        self._driver = driver

    async def read(self) -> Temperature:
        """Poll the tempdeck."""
        return await self._driver.get_temperature()


class TempdeckListener(WaitableListener[Temperature]):
    """Tempdeck state listener."""

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        interrupt_callback: types.InterruptCallback = None,
    ) -> None:
        """Constructor."""
        super().__init__(loop=loop)
        self._callback = interrupt_callback
        self._polled_data = Temperature(current=25, target=None)

    @property
    def state(self) -> Temperature:
        return self._polled_data

    def on_poll(self, result: Temperature) -> None:
        """On new poll."""
        self._polled_data = result
        return super().on_poll(result)

    def on_error(self, exc: Exception) -> None:
        """On error."""
        if self._callback:
            self._callback(str(exc))



"""Drop tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel
from typing import Optional, Type
from typing_extensions import Literal

from .pipetting_common import BasePipettingData
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest

DropTipCommandType = Literal["dropTip"]


class DropTipData(BasePipettingData):
    """Data required to drop a tip in a specific well."""

    pass


class DropTipResult(BaseModel):
    """Result data from the execution of a DropTipRequest."""

    pass


class DropTipImplementation(AbstractCommandImpl[DropTipData, DropTipResult]):
    """Drop tip command implementation."""

    async def execute(self, data: DropTipData) -> DropTipResult:
        """Move to and drop a tip using the requested pipette."""
        await self._pipetting.drop_tip(
            pipette_id=data.pipetteId,
            labware_id=data.labwareId,
            well_name=data.wellName,
        )

        return DropTipResult()


class DropTip(BaseCommand[DropTipData, DropTipResult]):
    """Drop tip command model."""

    commandType: DropTipCommandType = "dropTip"
    data: DropTipData
    result: Optional[DropTipResult]

    _ImplementationCls: Type[DropTipImplementation] = DropTipImplementation


class DropTipRequest(BaseCommandRequest[DropTipData]):
    """Drop tip command creation request model."""

    commandType: DropTipCommandType = "dropTip"
    data: DropTipData

    _CommandCls: Type[DropTip] = DropTip



import serial  # type: ignore
from serial.tools import list_ports  # type: ignore
import contextlib
import logging

log = logging.getLogger(__name__)

RECOVERY_TIMEOUT = 10
DEFAULT_SERIAL_TIMEOUT = 5
DEFAULT_WRITE_TIMEOUT = 30


class SerialNoResponse(Exception):
    pass


def get_ports_by_name(device_name):
    """Returns all serial devices with a given name"""
    filtered_devices = filter(
        lambda device: device_name in device[1], list_ports.comports()
    )
    device_ports = [device[0] for device in filtered_devices]
    return device_ports


def get_port_by_VID(vid):
    """Returns first serial device with a given VID"""
    for d in list_ports.comports():
        if d.vid == vid:
            return d[0]


@contextlib.contextmanager
def serial_with_temp_timeout(serial_connection, timeout):
    """Implements a temporary timeout for a serial connection"""
    saved_timeout = serial_connection.timeout
    if timeout is not None:
        serial_connection.timeout = timeout
    try:
        yield serial_connection
    finally:
        serial_connection.timeout = saved_timeout


def _parse_serial_response(response, ack):
    if ack in response:
        parsed_response = response.split(ack)[0]
        return parsed_response.strip()
    else:
        return None


def clear_buffer(serial_connection):
    serial_connection.reset_input_buffer()


def _write_to_device_and_return(cmd, ack, device_connection, tag=None):
    """Writes to a serial device.
    - Formats command
    - Wait for ack return
    - return parsed response"""

    if not tag:
        tag = device_connection.port

    encoded_write = cmd.encode()
    encoded_ack = ack.encode()
    log.debug(f"{tag}: Write -> {encoded_write}")
    device_connection.write(encoded_write)
    response = device_connection.read_until(encoded_ack)
    log.debug(f"{tag}: Read <- {response}")
    if encoded_ack not in response:
        log.warning(f"{tag}: timed out after {device_connection.timeout}")
        raise SerialNoResponse(
            "No response from serial port after {} second(s)".format(
                device_connection.timeout
            )
        )
    clean_response = _parse_serial_response(response, encoded_ack)
    if clean_response:
        return clean_response.decode()
    return ""


def _connect(port_name, baudrate):
    ser = serial.serial_for_url(
        url=port_name, baudrate=baudrate, timeout=DEFAULT_SERIAL_TIMEOUT
    )
    log.debug(ser)
    return ser


def _attempt_command_recovery(command, ack, serial_conn, tag=None):
    """Recovery after following a failed write_and_return() attempt"""
    if not tag:
        tag = serial_conn.port
    with serial_with_temp_timeout(serial_conn, RECOVERY_TIMEOUT) as device:
        response = _write_to_device_and_return(command, ack, device, tag=tag)
    if response is None:
        log.debug(f"{tag}: No valid response during _attempt_command_recovery")
        raise RuntimeError(
            "Recovery attempted - no valid serial response "
            "for command: {} in {} seconds".format(command.encode(), RECOVERY_TIMEOUT)
        )
    return response


def write_and_return(
    command, ack, serial_connection, timeout=DEFAULT_WRITE_TIMEOUT, tag=None
):
    """Write a command and return the response"""
    clear_buffer(serial_connection)
    with serial_with_temp_timeout(serial_connection, timeout) as device_connection:
        response = _write_to_device_and_return(command, ack, device_connection, tag)
    return response


def connect(device_name=None, port=None, baudrate=115200):
    """
    Creates a serial connection
    :param device_name: defaults to 'Smoothieboard'
    :param baudrate: integer frequency for serial communication
    :return: serial.Serial connection
    """
    if not port:
        port = get_ports_by_name(device_name=device_name)[0]
    log.debug("Device name: {}, Port: {}".format(device_name, port))
    return _connect(port_name=port, baudrate=baudrate)



from .helpers import stringify_location
from . import types as command_types

from typing import Union, Sequence, Optional, TYPE_CHECKING

from opentrons.protocols.api_support.util import FlowRates

if TYPE_CHECKING:
    from opentrons.protocol_api.instrument_context import InstrumentContext
    from opentrons.protocol_api.labware import Well
    from opentrons.types import Location


Apiv2Locations = Sequence[Union["Location", "Well"]]
Apiv2Instruments = Sequence["InstrumentContext"]


def combine_locations(location: Sequence) -> str:
    if len(location) > 1:
        loc1 = stringify_location(location[0])
        loc2 = stringify_location(location[1])
        return f"{loc1} and {loc2}"
    elif len(location) == 1:
        loc1 = stringify_location(location[0])
        return f"{loc1}"
    else:
        return ""


def paired_aspirate(
    instruments: Apiv2Instruments,
    volume: float,
    locations: Apiv2Locations,
    rate: float,
    pub_type: str,
) -> command_types.AspirateCommand:
    loc_text = combine_locations(locations)
    flow_rate = min(
        rate * FlowRates(instr._implementation).aspirate for instr in instruments
    )
    text_type = f"{pub_type}: Aspirating "
    text_content = f"{volume} uL from {loc_text} at {flow_rate} uL/sec"
    text = text_type + text_content
    return {
        "name": command_types.ASPIRATE,
        "payload": {
            "instruments": instruments,
            "volume": volume,
            "locations": locations,
            "rate": rate,
            "text": text,
        },
    }


def paired_dispense(
    instruments: Apiv2Instruments,
    volume: float,
    locations: Apiv2Locations,
    rate: float,
    pub_type: str,
) -> command_types.DispenseCommand:
    loc_text = combine_locations(locations)
    flow_rate = min(
        rate * FlowRates(instr._implementation).dispense for instr in instruments
    )
    text_type = f"{pub_type}: Dispensing "
    text_content = f"{volume} uL into {loc_text} at {flow_rate} uL/sec"
    text = text_type + text_content
    return {
        "name": command_types.DISPENSE,
        "payload": {
            "instruments": instruments,
            "volume": volume,
            "locations": locations,
            "rate": rate,
            "text": text,
        },
    }


def paired_mix(
    instruments: Apiv2Instruments,
    locations: Optional[Apiv2Locations],
    repetitions: int,
    volume: float,
    pub_type: str,
) -> command_types.MixCommand:
    text_type = f"{pub_type}: Mixing "
    text_content = "{repetitions} times with a volume of {volume} ul"
    text = text_type + text_content
    return {
        "name": command_types.MIX,
        "payload": {
            "instruments": instruments,
            "locations": locations,
            "volume": volume,
            "repetitions": repetitions,
            "text": text,
        },
    }


def paired_blow_out(
    instruments: Apiv2Instruments, locations: Optional[Apiv2Locations], pub_type: str
) -> command_types.BlowOutCommand:
    text = f"{pub_type}: Blowing out"

    if locations is not None:
        location_text = combine_locations(locations)
        text += f" at {location_text}"

    return {
        "name": command_types.BLOW_OUT,
        "payload": {"instruments": instruments, "locations": locations, "text": text},
    }


def paired_touch_tip(
    instruments: Apiv2Instruments, locations: Optional[Apiv2Locations], pub_type: str
) -> command_types.TouchTipCommand:
    text = f"{pub_type}: Touching tip"

    if locations is not None:
        location_text = combine_locations(locations)
        text += f" at {location_text}"
    return {
        "name": command_types.TOUCH_TIP,
        "payload": {"instruments": instruments, "locations": locations, "text": text},
    }


def air_gap() -> command_types.AirGapCommand:
    text = "Air gap"
    return {"name": command_types.AIR_GAP, "payload": {"text": text}}


def return_tip() -> command_types.ReturnTipCommand:
    text = "Returning tip"
    return {"name": command_types.RETURN_TIP, "payload": {"text": text}}


def paired_pick_up_tip(
    instruments: Apiv2Instruments, locations: Apiv2Locations, pub_type: str
) -> command_types.PickUpTipCommand:
    location_text = combine_locations(locations)
    text = f"{pub_type}: Picking up tip from {location_text}"
    return {
        "name": command_types.PICK_UP_TIP,
        "payload": {"instruments": instruments, "locations": locations, "text": text},
    }


def paired_drop_tip(
    instruments: Apiv2Instruments, locations: Apiv2Locations, pub_type: str
) -> command_types.DropTipCommand:
    location_text = combine_locations(locations)
    text = f"{pub_type}: Dropping tip into {location_text}"
    return {
        "name": command_types.DROP_TIP,
        "payload": {"instruments": instruments, "locations": locations, "text": text},
    }


def paired_move_to(
    instruments: Apiv2Instruments, locations: Apiv2Locations, pub_type: str
) -> command_types.MoveToCommand:
    location_text = combine_locations(locations)
    text = f"{pub_type}: Moving to {location_text}"
    return {
        "name": command_types.MOVE_TO,
        "payload": {"instruments": instruments, "locations": locations, "text": text},
    }



from typing import Optional

from opentrons.drivers.thermocycler.abstract import AbstractThermocyclerDriver
from opentrons.drivers.types import Temperature, PlateTemperature, ThermocyclerLidStatus


class SimulatingDriver(AbstractThermocyclerDriver):
    DEFAULT_TEMP = 23

    def __init__(self):
        self._ramp_rate: Optional[float] = None
        self._lid_status = ThermocyclerLidStatus.OPEN
        self._lid_temperature = Temperature(current=self.DEFAULT_TEMP, target=None)
        self._plate_temperature = PlateTemperature(
            current=self.DEFAULT_TEMP, target=None, hold=None
        )

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    async def is_connected(self) -> bool:
        return True

    async def open_lid(self) -> None:
        self._lid_status = ThermocyclerLidStatus.OPEN

    async def close_lid(self) -> None:
        self._lid_status = ThermocyclerLidStatus.CLOSED

    async def get_lid_status(self) -> ThermocyclerLidStatus:
        return self._lid_status

    async def get_lid_temperature(self) -> Temperature:
        return self._lid_temperature

    async def set_plate_temperature(
        self,
        temp: float,
        hold_time: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        self._plate_temperature.target = temp
        self._plate_temperature.current = temp
        self._plate_temperature.hold = hold_time

    async def get_plate_temperature(self) -> PlateTemperature:
        return self._plate_temperature

    async def set_ramp_rate(self, ramp_rate: float) -> None:
        self._ramp_rate = ramp_rate

    async def set_lid_temperature(self, temp: float) -> None:
        """Set the lid temperature in deg Celsius"""
        self._lid_temperature.target = temp
        self._lid_temperature.current = temp

    async def deactivate_lid(self):
        self._lid_temperature.target = None
        self._lid_temperature.current = self.DEFAULT_TEMP

    async def deactivate_block(self):
        self._plate_temperature.target = None
        self._plate_temperature.current = self.DEFAULT_TEMP
        self._plate_temperature.hold = None
        self._ramp_rate = None

    async def deactivate_all(self):
        await self.deactivate_lid()
        await self.deactivate_block()

    async def get_device_info(self):
        return {
            "serial": "dummySerialTC",
            "model": "dummyModelTC",
            "version": "dummyVersionTC",
        }

    async def enter_programming_mode(self):
        pass



"""Pick up tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel
from typing import Optional, Type
from typing_extensions import Literal

from .pipetting_common import BasePipettingData
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest


PickUpTipCommandType = Literal["pickUpTip"]


class PickUpTipData(BasePipettingData):
    """Data needed to move a pipette to a specific well."""

    pass


class PickUpTipResult(BaseModel):
    """Result data from the execution of a PickUpTip."""

    pass


class PickUpTipImplementation(AbstractCommandImpl[PickUpTipData, PickUpTipResult]):
    """Pick up tip command implementation."""

    async def execute(self, data: PickUpTipData) -> PickUpTipResult:
        """Move to and pick up a tip using the requested pipette."""
        await self._pipetting.pick_up_tip(
            pipette_id=data.pipetteId,
            labware_id=data.labwareId,
            well_name=data.wellName,
        )

        return PickUpTipResult()


class PickUpTip(BaseCommand[PickUpTipData, PickUpTipResult]):
    """Pick up tip command model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    data: PickUpTipData
    result: Optional[PickUpTipResult]

    _ImplementationCls: Type[PickUpTipImplementation] = PickUpTipImplementation


class PickUpTipRequest(BaseCommandRequest[PickUpTipData]):
    """Pick up tip command creation request model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    data: PickUpTipData

    _CommandCls: Type[PickUpTip] = PickUpTip



"""Motion planning base interfaces."""
from dataclasses import dataclass
from enum import Enum, auto as auto_enum_value
from typing import Optional
from typing_extensions import final

from opentrons.types import Point
from opentrons.hardware_control.types import CriticalPoint


@dataclass(frozen=True)
@final
class Waypoint:
    """Motion waypoint with position and optional pipette critical point."""

    position: Point
    critical_point: Optional[CriticalPoint] = None


@final
class MoveType(Enum):
    """Arc or direct movement type.

    - GENERAL_ARC: an arc movement between two unrelated locations
    - IN_LABWARE_ARC: an arc movement between two locations in the same labware
    - DIRECT: a direct movement between two locations
    """

    GENERAL_ARC = auto_enum_value()
    IN_LABWARE_ARC = auto_enum_value()
    DIRECT = auto_enum_value()



import abc


class DeckItem(abc.ABC):
    @property  # type: ignore
    @abc.abstractmethod
    def highest_z(self):
        pass

    @highest_z.setter  # type: ignore
    @abc.abstractmethod
    def highest_z(self, new_z: float):
        pass

    @property
    @abc.abstractmethod
    def separate_calibration(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def load_name(self) -> str:
        pass



# noqa: D100
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional

from opentrons.protocols.geometry.well_geometry import WellGeometry
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient

from .types import Location, Point

if TYPE_CHECKING:
    from .labware import Labware


class Well:  # noqa: D101
    def __init__(
        self,
        engine_client: ProtocolEngineClient,
        labware: Labware,
        well_name: str,
    ) -> None:
        """Initialize a Well API provider.

        You should not need to call this constructor yourself. The system will
        create `Well`s for you when you call :py:meth:`load_labware`.

        Args:
            engine_client: A client to access protocol state.
            labware: The well's parent Labware instance.
            well_name: The unique name of the well inside its parent labware.
        """
        self._engine_client = engine_client
        self._labware = labware
        self._well_name = well_name
        self._well_definition = self._engine_client.state.labware.get_well_definition(
            labware_id=labware.labware_id, well_name=well_name
        )

    # TODO(mc, 2021-04-22): remove this property; it's redundant and
    # unlikely to be used by PAPI users
    @property
    def api_version(self) -> Any:  # noqa: D102
        raise NotImplementedError()

    @property
    def parent(self) -> Labware:  # noqa: D102
        return self._labware

    @property
    def has_tip(self) -> bool:  # noqa: D102
        raise NotImplementedError()

    @has_tip.setter
    def has_tip(self, value: bool) -> None:
        raise NotImplementedError()

    @property
    def max_volume(self) -> float:  # noqa: D102
        return self._well_definition.totalLiquidVolume

    # TODO(mc, 2021-04-22): explore collapsing WellGeometry into Well
    @property
    def geometry(self) -> WellGeometry:  # noqa: D102
        raise NotImplementedError()

    @property
    def diameter(self) -> Optional[float]:  # noqa: D102
        return self._well_definition.diameter

    @property
    def length(self) -> Optional[float]:  # noqa: D102
        raise NotImplementedError()

    @property
    def width(self) -> Optional[float]:  # noqa: D102
        raise NotImplementedError()

    @property
    def depth(self) -> float:  # noqa: D102
        return self._well_definition.depth

    @property
    def display_name(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def well_name(self) -> str:  # noqa: D102
        return self._well_name

    def top(self, z: float = 0.0) -> Location:  # noqa: D102
        raise NotImplementedError()

    def bottom(self, z: float = 0.0) -> Location:  # noqa: D102
        raise NotImplementedError()

    def center(self) -> Location:  # noqa: D102
        raise NotImplementedError()

    def from_center_cartesian(  # noqa: D102
        self,
        x: float,
        y: float,
        z: float,
    ) -> Point:
        raise NotImplementedError()

    def __repr__(self) -> str:  # noqa: D105
        return f"Well:{self._well_name}<Labware:{self.parent}>"

    def __eq__(self, other: object) -> bool:
        """Compare for object equality.

        Checks that other object is a `Well` and belongs to the same labware.
        """
        return (
            isinstance(other, Well)
            and self.well_name == other.well_name
            and self.parent == other.parent
        )

    def __hash__(self) -> int:  # noqa: D105
        raise NotImplementedError()



from .driver import ThermocyclerDriver
from .simulator import SimulatingDriver
from .abstract import AbstractThermocyclerDriver


__all__ = ["ThermocyclerDriver", "SimulatingDriver", "AbstractThermocyclerDriver"]



"""An emulation of the opentrons magnetic module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional
from opentrons.drivers.mag_deck.driver import GCODE
from opentrons.hardware_control.emulation.parser import Parser, Command
from .abstract_emulator import AbstractEmulator

logger = logging.getLogger(__name__)


SERIAL = "magnetic_emulator"
MODEL = "mag_deck_v20"
VERSION = "2.0.0"


class MagDeckEmulator(AbstractEmulator):
    """Magdeck emulator"""

    def __init__(self, parser: Parser) -> None:
        self.reset()
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = " ".join(r for r in results if r)
        return None if not joined else joined

    def reset(self):
        self.height: float = 0
        self.position: float = 0

    def _handle(self, command: Command) -> Optional[str]:
        """Handle a command."""
        logger.info(f"Got command {command}")
        if command.gcode == GCODE.HOME:
            self.height = 0
        elif command.gcode == GCODE.MOVE:
            position = command.params["Z"]
            assert isinstance(position, float), f"invalid position '{position}'"
            self.position = position
        elif command.gcode == GCODE.PROBE_PLATE:
            self.height = 45
        elif command.gcode == GCODE.GET_PLATE_HEIGHT:
            return f"height:{self.height}"
        elif command.gcode == GCODE.GET_CURRENT_POSITION:
            return f"Z:{self.position}"
        elif command.gcode == GCODE.DEVICE_INFO:
            return f"serial:{SERIAL} model:{MODEL} version:{VERSION}"
        elif command.gcode == GCODE.PROGRAMMING_MODE:
            pass
        return None



import logging

from opentrons import types
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.context.protocol_api.protocol_context import (
    ProtocolContextImplementation,
)
from opentrons.protocols.context.simulator.instrument_context import (
    InstrumentContextSimulation,
)


logger = logging.getLogger(__name__)


class ProtocolContextSimulation(ProtocolContextImplementation):
    def load_instrument(
        self, instrument_name: str, mount: types.Mount, replace: bool
    ) -> AbstractInstrument:
        """Create a simulating instrument context."""
        instr = self._instruments[mount]
        if instr:
            # There's already an instrument on this mount.
            if not replace:
                # If not replacing then error.
                raise RuntimeError(
                    f"Instrument already present in {mount.name.lower()} "
                    f"mount: {instr.get_instrument_name()}"
                )
            elif instr.get_instrument_name() == instrument_name:
                # Replacing with the exact same instrument name. Just return the
                # existing instrument instance.
                return instr

        attached = {
            att_mount: instr.get("name", None)
            for att_mount, instr in self._hw_manager.hardware.attached_instruments.items()  # noqa: E501
            if instr
        }
        attached[mount] = instrument_name
        self._hw_manager.hardware.cache_instruments(attached)

        new_instr = InstrumentContextSimulation(
            protocol_interface=self,
            pipette_dict=self._hw_manager.hardware.get_attached_instruments()[mount],
            mount=mount,
            instrument_name=instrument_name,
            api_version=self._api_version,
        )
        self._instruments[mount] = new_instr
        logger.info(f"Instrument {new_instr} loaded")
        return new_instr



import argparse
from typing import Tuple, cast

from opentrons.hardware_control import API
from opentrons.drivers.smoothie_drivers import SmoothieDriver


def root_argparser(description: str = None):
    parse = argparse.ArgumentParser(description=description)
    parse.add_argument(
        "-p", "--port", help="serial port of the smoothie", default="", type=str
    )
    return parse


async def build_driver(port: str = None) -> Tuple[API, SmoothieDriver]:
    hardware = await API.build_hardware_controller(port=port)
    driver = cast(SmoothieDriver, hardware._backend._smoothie_driver)
    return hardware, driver



from abc import ABC, abstractmethod
from typing import Optional, Dict

from opentrons.drivers.types import Temperature, ThermocyclerLidStatus, PlateTemperature


class AbstractThermocyclerDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """Connect to thermocycler"""
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from thermocycler"""
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        """Check connection"""
        ...

    @abstractmethod
    async def open_lid(self) -> None:
        """Send open lid command"""
        ...

    @abstractmethod
    async def close_lid(self) -> None:
        """Send close lid command"""
        ...

    @abstractmethod
    async def get_lid_status(self) -> ThermocyclerLidStatus:
        """Send get lid status command"""
        ...

    @abstractmethod
    async def set_lid_temperature(self, temp: float) -> None:
        """Set the lid temperature"""
        ...

    @abstractmethod
    async def get_lid_temperature(self) -> Temperature:
        """Send a get lid temperature command."""
        ...

    @abstractmethod
    async def set_plate_temperature(
        self,
        temp: float,
        hold_time: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        """Send set plate temperature command"""
        ...

    @abstractmethod
    async def get_plate_temperature(self) -> PlateTemperature:
        """Send a get plate temperature command."""
        ...

    @abstractmethod
    async def set_ramp_rate(self, ramp_rate: float) -> None:
        """Send a set ramp rate command"""
        ...

    @abstractmethod
    async def deactivate_all(self) -> None:
        """Send deactivate all command."""
        ...

    @abstractmethod
    async def deactivate_lid(self) -> None:
        """Send deactivate lid command"""
        ...

    @abstractmethod
    async def deactivate_block(self) -> None:
        """Send deactivate block command"""
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        """Send get device info command"""
        ...

    @abstractmethod
    async def enter_programming_mode(self) -> None:
        ...



"""Protocol engine state module."""

from .create_state_store import create_state_store
from .state import State, StateStore, StateView
from .commands import CommandState, CommandView
from .labware import LabwareState, LabwareView
from .pipettes import PipetteState, PipetteView, HardwarePipette, CurrentWell
from .geometry import GeometryView, TipGeometry
from .motion import MotionView, PipetteLocationData
from .actions import Action, PlayAction, PauseAction, StopAction, UpdateCommandAction

__all__ = [
    # top level store factory
    "create_state_store",
    # top level state value and interfaces
    "State",
    "StateStore",
    "StateView",
    # command state
    "CommandState",
    "CommandView",
    # labware state
    "LabwareState",
    "LabwareView",
    # pipette state
    "PipetteState",
    "PipetteView",
    "HardwarePipette",
    "CurrentWell",
    # computed geometry state
    "GeometryView",
    "TipGeometry",
    # computed motion state
    "MotionView",
    "PipetteLocationData",
    # actions
    "Action",
    "PlayAction",
    "PauseAction",
    "StopAction",
    "UpdateCommandAction",
]



"""Simulating ProtocolRunner factory."""

from opentrons.hardware_control import API as HardwareAPI
from opentrons.protocol_engine import create_protocol_engine
from .protocol_runner import ProtocolRunner


async def create_simulating_runner() -> ProtocolRunner:
    """Create a ProtocolRunner wired to a simulating HardwareAPI.

    Example:
        ```python
        from pathlib import Path
        from typing import List
        from opentrons.protocol_engine import Command
        from opentrons.protocol_runner import (
            ProtocolType,
            ProtocolFile,
            ProtocolRunner,
            create_simulating_runner,
        )

        protocol = ProtocolFile(
            protocol_type=ProtocolType.PYTHON,
            files=[Path("/path/to/protocol.py")],
        )
        runner: ProtocolRunner = await create_simulating_runner()
        commands: List[Command] = await runner.run(protocol)
        ```
    """
    simulating_hardware_api = await HardwareAPI.build_hardware_simulator()

    # TODO(mc, 2021-08-25): move initial home to protocol engine
    await simulating_hardware_api.home()

    # TODO(mc, 2021-08-25): this engine will not simulate pauses
    # https://github.com/Opentrons/opentrons/issues/8265
    protocol_engine = await create_protocol_engine(hardware_api=simulating_hardware_api)

    return ProtocolRunner(protocol_engine=protocol_engine)



"""Value objects and models representing protocol files."""
# TODO(mc, 2021-04-30): as these objects are fleshed out, pull in
# existing logic and models from:
#   - api/src/opentrons/protocols/types.py
#   - robot-server/robot_server/service/protocol/models.py
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List


class ProtocolFileType(str, Enum):
    """Type of a protocol file.

    Attributes:
        PYTHON: a Python protocol file or module
        JSON: a JSON protocol
    """

    PYTHON = "python"
    JSON = "json"


# TODO(mc, 2021-08-27): rename to ProtocolSource to better reflect
# the fact that a protocol need not be a single file
@dataclass(frozen=True)
class ProtocolFile:
    """A value object representing a protocol file (or files) on disk.

    Attributes:
        protocol_type: Whether the file is a JSON protocol or Python protocol.
        files: The list of files that make up the protocol.
    """

    # TODO(mc, 2021-08-27): `protocol_type` is a little redundant as a field name
    protocol_type: ProtocolFileType
    files: List[Path]



"""Move to well command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel
from typing import Optional, Type
from typing_extensions import Literal

from .pipetting_common import BasePipettingData
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest


MoveToWellCommandType = Literal["moveToWell"]


class MoveToWellData(BasePipettingData):
    """Data required to move a pipette to a specific well."""

    pass


class MoveToWellResult(BaseModel):
    """Result data from the execution of a MoveToWell command."""

    pass


class MoveToWellImplementation(AbstractCommandImpl[MoveToWellData, MoveToWellResult]):
    """Move to well command implementation."""

    async def execute(self, data: MoveToWellData) -> MoveToWellResult:
        """Move the requested pipette to the requested well."""
        await self._movement.move_to_well(
            pipette_id=data.pipetteId,
            labware_id=data.labwareId,
            well_name=data.wellName,
        )

        return MoveToWellResult()


class MoveToWell(BaseCommand[MoveToWellData, MoveToWellResult]):
    """Move to well command model."""

    commandType: MoveToWellCommandType = "moveToWell"
    data: MoveToWellData
    result: Optional[MoveToWellResult]

    _ImplementationCls: Type[MoveToWellImplementation] = MoveToWellImplementation


class MoveToWellRequest(BaseCommandRequest[MoveToWellData]):
    """Move to well command creation request model."""

    commandType: MoveToWellCommandType = "moveToWell"
    data: MoveToWellData

    _CommandCls: Type[MoveToWell] = MoveToWell



from typing import NamedTuple


class APIVersion(NamedTuple):
    major: int
    minor: int

    @classmethod
    def from_string(cls, inp: str) -> "APIVersion":
        parts = inp.split(".")
        if len(parts) != 2:
            raise ValueError(inp)
        intparts = [int(p) for p in parts]

        return cls(major=intparts[0], minor=intparts[1])

    def __str__(self):
        return f"{self.major}.{self.minor}"



import logging

from opentrons.protocol_api.contexts import ProtocolContext
from opentrons.protocols.execution.execute_python import run_python
from opentrons.protocols.execution.json_dispatchers import (
    pipette_command_map,
    temperature_module_command_map,
    magnetic_module_command_map,
    thermocycler_module_command_map,
)
from opentrons.protocols.execution import execute_json_v4, execute_json_v3

from opentrons.protocols.types import PythonProtocol, Protocol
from opentrons.protocols.api_support.types import APIVersion

MODULE_LOG = logging.getLogger(__name__)


def run_protocol(protocol: Protocol, context: ProtocolContext):
    """Run a protocol.

    :param protocol: The :py:class:`.protocols.types.Protocol` to execute
    :param context: The context to use.
    """
    if isinstance(protocol, PythonProtocol):
        if protocol.api_level >= APIVersion(2, 0):
            run_python(protocol, context)
        else:
            raise RuntimeError(f"Unsupported python API version: {protocol.api_level}")
    else:
        if protocol.contents["schemaVersion"] == 3:
            ins = execute_json_v3.load_pipettes_from_json(context, protocol.contents)
            lw = execute_json_v3.load_labware_from_json_defs(context, protocol.contents)
            execute_json_v3.dispatch_json(context, protocol.contents, ins, lw)
        elif protocol.contents["schemaVersion"] in [4, 5]:
            # reuse the v3 fns for loading labware and pipettes
            # b/c the v4 protocol has no changes for these keys
            ins = execute_json_v3.load_pipettes_from_json(context, protocol.contents)

            modules = execute_json_v4.load_modules_from_json(context, protocol.contents)

            lw = execute_json_v4.load_labware_from_json_defs(
                context, protocol.contents, modules
            )
            execute_json_v4.dispatch_json(
                context,
                protocol.contents,
                ins,
                lw,
                modules,
                pipette_command_map,
                magnetic_module_command_map,
                temperature_module_command_map,
                thermocycler_module_command_map,
            )
        else:
            raise RuntimeError(
                f"Unsupported JSON protocol schema: {protocol.schema_version}"
            )



""" opentrons.calibration_storage.helpers: various miscellaneous
functions

This module has functions that you can import to save robot or
labware calibration to its designated file location.
"""
import json
from typing import Union, List, Dict, TYPE_CHECKING, cast
from dataclasses import is_dataclass, asdict


from hashlib import sha256

from . import types as local_types

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
    from opentrons_shared_data.pipette.dev_types import LabwareUri


DictionaryFactoryType = Union[List, Dict]


def dict_filter_none(data: DictionaryFactoryType) -> Dict:
    """
    Helper function to filter out None keys from a dataclass
    before saving to file.
    """
    return dict(item for item in data if item[1] is not None)


def convert_to_dict(obj) -> Dict:
    # The correct way to type this is described here:
    # https://github.com/python/mypy/issues/6568
    # Unfortnately, since it's not currently supported I have an
    # assert check instead.
    assert is_dataclass(obj), "This function is intended for dataclasses only"
    return asdict(obj, dict_factory=dict_filter_none)


def hash_labware_def(labware_def: "LabwareDefinition") -> str:
    """
    Helper function to take in a labware definition and return
    a hashed string of key elemenets from the labware definition
    to make it a unique identifier.

    :param labware_def: Full labware definitino
    :returns: sha256 string
    """
    # remove keys that do not affect run
    blocklist = ["metadata", "brand", "groups"]
    def_no_metadata = {k: v for k, v in labware_def.items() if k not in blocklist}
    sorted_def_str = json.dumps(def_no_metadata, sort_keys=True, separators=(",", ":"))
    return sha256(sorted_def_str.encode("utf-8")).hexdigest()


def details_from_uri(uri: str, delimiter="/") -> local_types.UriDetails:
    """
    Unpack a labware URI to get the namespace, loadname and version
    """
    if uri:
        info = uri.split(delimiter)
        return local_types.UriDetails(
            namespace=info[0], load_name=info[1], version=int(info[2])
        )
    else:
        # Here we are assuming that the 'uri' passed in is actually
        # the loadname, though sometimes it may be an empty string.
        return local_types.UriDetails(namespace="", load_name=uri, version=1)


def uri_from_details(
    namespace: str, load_name: str, version: Union[str, int], delimiter="/"
) -> "LabwareUri":
    """Build a labware URI from its details.

    A labware URI is a string that uniquely specifies a labware definition.

    :returns str: The URI.
    """
    return cast("LabwareUri", f"{namespace}{delimiter}{load_name}{delimiter}{version}")


def uri_from_definition(definition: "LabwareDefinition", delimiter="/") -> "LabwareUri":
    """Build a labware URI from its definition.

    A labware URI is a string that uniquely specifies a labware definition.

    :returns str: The URI.
    """
    return uri_from_details(
        namespace=definition["namespace"],
        load_name=definition["parameters"]["loadName"],
        version=definition["version"],
        delimiter=delimiter,
    )



"""The interface that implements ProtocolContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Dict, Optional, Union

from opentrons import types
from opentrons.hardware_control import SynchronousAdapter, ThreadManager
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.protocols.geometry.deck import Deck
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocols.geometry.module_geometry import ModuleGeometry
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.api_support.util import AxisMaxSpeeds, HardwareManager
from opentrons.protocols.context.labware import AbstractLabware
from opentrons_shared_data.labware.dev_types import LabwareDefinition


InstrumentDict = Dict[types.Mount, Optional[AbstractInstrument]]


@dataclass(frozen=True)
class LoadModuleResult:
    """The result of load_module"""

    type: ModuleType
    geometry: ModuleGeometry
    module: SynchronousAdapter


class AbstractProtocol(ABC):
    @abstractmethod
    def get_bundled_data(self) -> Dict[str, bytes]:
        """Get a mapping of name to contents"""
        ...

    @abstractmethod
    def get_bundled_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        ...

    @abstractmethod
    def get_extra_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        ...

    @abstractmethod
    def cleanup(self) -> None:
        ...

    @abstractmethod
    def get_max_speeds(self) -> AxisMaxSpeeds:
        ...

    @abstractmethod
    def get_hardware(self) -> HardwareManager:
        ...

    @abstractmethod
    def connect(self, hardware: Union[ThreadManager, SynchronousAdapter]) -> None:
        ...

    @abstractmethod
    def disconnect(self) -> None:
        ...

    @abstractmethod
    def is_simulating(self) -> bool:
        ...

    @abstractmethod
    def load_labware_from_definition(
        self,
        labware_def: LabwareDefinition,
        location: types.DeckLocation,
        label: Optional[str],
    ) -> AbstractLabware:
        ...

    @abstractmethod
    def load_labware(
        self,
        load_name: str,
        location: types.DeckLocation,
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
    ) -> AbstractLabware:
        ...

    @abstractmethod
    def load_module(
        self,
        module_name: str,
        location: Optional[types.DeckLocation],
        configuration: Optional[str],
    ) -> Optional[LoadModuleResult]:
        ...

    @abstractmethod
    def get_loaded_modules(self) -> Dict[int, LoadModuleResult]:
        ...

    @abstractmethod
    def load_instrument(
        self, instrument_name: str, mount: types.Mount, replace: bool
    ) -> AbstractInstrument:
        ...

    @abstractmethod
    def get_loaded_instruments(self) -> InstrumentDict:
        ...

    @abstractmethod
    def pause(self, msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def resume(self) -> None:
        ...

    @abstractmethod
    def comment(self, msg: str) -> None:
        ...

    @abstractmethod
    def delay(self, seconds: float, msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def get_deck(self) -> Deck:
        ...

    @abstractmethod
    def get_fixed_trash(self) -> DeckItem:
        ...

    @abstractmethod
    def set_rail_lights(self, on: bool) -> None:
        ...

    @abstractmethod
    def get_rail_lights_on(self) -> bool:
        ...

    @abstractmethod
    def door_closed(self) -> bool:
        ...

    @abstractmethod
    def get_last_location(
        self,
        mount: Optional[types.Mount] = None,
    ) -> Optional[types.Location]:
        ...

    @abstractmethod
    def set_last_location(
        self,
        location: Optional[types.Location],
        mount: Optional[types.Mount] = None,
    ) -> None:
        ...



from typing import cast, List, Optional, Tuple
from opentrons.types import Point
from opentrons.protocol_api import labware, InstrumentContext, ProtocolContext

from opentrons.hardware_control.modules.types import ModuleType
from opentrons.protocols.geometry.module_geometry import ModuleGeometry


def _get_parent_slot_and_position(
    labware_obj: labware.Labware,
) -> Tuple[str, Optional[Point]]:
    if isinstance(labware_obj.parent, (ModuleGeometry)):
        return (cast(str, labware_obj.parent.parent), labware_obj.parent.labware_offset)
    else:
        return (cast(str, labware_obj.parent), None)


class Container:
    def __init__(
        self,
        container: labware.Labware,
        instruments: List[InstrumentContext],
        context: ProtocolContext,
    ) -> None:
        self._container = container
        self._context = context
        self.id = id(container)
        self.labware_offset_from_slot = 0

        # will be labware's load name or label
        self.name = container.name
        # type must be load_name so client can load correct definition
        self.type = container.load_name
        slot, position = _get_parent_slot_and_position(container)
        self.slot = slot
        self.position = position
        self.is_legacy = False
        self.is_tiprack = container.is_tiprack
        self.definition_hash = labware.get_labware_hash_with_parent(container)
        self.instruments = [
            Instrument(instrument, [], self._context) for instrument in instruments
        ]


class Instrument:
    def __init__(
        self,
        instrument: InstrumentContext,
        containers: List[labware.Labware],
        context: ProtocolContext,
    ) -> None:
        containers = containers or []
        self._instrument = instrument
        self._context = context

        self.id = id(instrument)
        # The name element here is actually the pipette model for historical
        # reasons
        self.name = instrument.model
        self.model_name = instrument.name
        self.channels = instrument.channels
        self.mount = instrument.mount
        self.containers = [
            Container(container, [], self._context) for container in containers
        ]
        self.tip_racks = [
            Container(container, [], self._context)
            for container in instrument.tip_racks
        ]
        if context:
            self.tip_racks.extend(
                [c for c in self.containers if c._container.is_tiprack]
            )
        self.requested_as = instrument.requested_as


class Module:
    def __init__(self, module: ModuleGeometry, context: ProtocolContext) -> None:
        self.id = id(module)
        _type_lookup = {
            ModuleType.MAGNETIC: "magdeck",
            ModuleType.TEMPERATURE: "tempdeck",
            ModuleType.THERMOCYCLER: "thermocycler",
        }
        self.name = _type_lookup[module.module_type]
        self.model = module.model.value
        self.slot = module.parent
        self._context = context
